from math import cos, sin, pi

from ase import Atoms
from ase.calculators.tip4p import TIP4P, epsilon0, sigma0, rOH, angleHOH
from ase.calculators.qmmm import EIQMMM, LJInteractions, Embedding
from ase.constraints import FixBondLengths
from ase.optimize import LBFGS
from ase.optimize.precon import PreconLBFGS, Exp
from gpaw import GPAW

r = rOH
a = angleHOH / 180 * pi

interaction = LJInteractions({('O', 'O'): (epsilon0, sigma0)})

for selection in [[0, 1, 2], [3, 4, 5]]:
    name = ''.join(str(i) for i in selection)
    dimer = Atoms('OH2OH2',
                  [(0, 0, 0),
                   (-r * cos(a / 2), r * sin(a / 2), 0),
                   (-r * cos(a / 2), -r * sin(a / 2), 0),
                   (0, 0, 0),
                   (-r * cos(a), 0, r * sin(a)),
                   (-r, 0, 0)])
    dimer.positions[3:, 0] += 2.8
    dimer.constraints = FixBondLengths(
        [((selection[i] + 3) % 6, (selection[i - 1] + 3) % 6)
         for i in range(3)])

    dimer.calc = EIQMMM(selection,
                        GPAW(txt=name + '.txt', h=0.16),
                        TIP4P(),
                        interaction,
                        vacuum=4,
                        embedding=Embedding(rc=0.2, rc2=20, width=1),
                        output=name + '.out')
    opt = LBFGS(dimer, trajectory=name + '.traj')
    opt.run(0.02)

    monomer = dimer[selection]
    monomer.center(vacuum=4)
    # Grrr.  PreconLBFGS breaks the symmetry!  Some one should fix that.
    monomer.calc = GPAW(txt=name + 'M.txt', h=0.16, symmetry='off')
    opt = PreconLBFGS(monomer, precon=Exp(A=3), trajectory=name + 'M.traj')
    opt.run(0.02)
    e0 = monomer.get_potential_energy()
    be = dimer.get_potential_energy() - e0
    d = dimer.get_distance(0, 3)
    print(name, be, d)
    if name == '012':
        assert abs(be - -0.287) < 0.002
        assert abs(d - 2.76) < 0.02
    else:
        assert abs(be - -0.316) < 0.002
        assert abs(d - 2.67) < 0.02



from gpaw.wannier.w90 import read_wout_all
from myqueue.task import task
import numpy as np


def create_tasks():
    t1 = task('GaAs.py')
    t2 = task('GaAs_wannier.py', deps=t1)
    t3 = task('Fe.py', cores=8)
    t4 = task('Fe_wannier.py', nodename='xeon8_48', tmax='1h', deps=t3)
    t5 = task('agts.py', deps=[t2, t4])
    return [t1, t2, t3, t4, t5]


if __name__ == '__main__':
    with open('GaAs.wout') as fd:
        dct = read_wout_all(fd)
    x, y, z = dct['centers'].sum(axis=0)
    w = dct['spreads'].sum()
    a = 5.68
    assert abs(np.array([x, y, z, w]) - [a, a, a, 4.14]).max() < 0.01

    with open('Fe.wout') as fd:
        dct = read_wout_all(fd)
    x, y, z = dct['centers'].sum(axis=0)
    w = dct['spreads'].sum()
    assert abs(np.array([x, y, z, w]) - [0, 0, 0, 14.49]).max() < 0.01



from ase.parallel import paropen
from ase.units import Hartree
from gpaw.xc.rpa import RPACorrelation
from gpaw.xc.fxc import FXCCorrelation
from gpaw.mpi import world

fxc0 = FXCCorrelation('CO.ralda.pbe_wfcs_O.gpw',
                      xc='rAPBE',
                      txt='CO.ralda_04_O_rapbe.txt',
                      wcomm=world.size)

E0_i = fxc0.calculate(ecut=400)

f = paropen('CO.ralda_rapbe_O.dat', 'w')
for ecut, E0 in zip(fxc0.ecut_i, E0_i):
    print(ecut * Hartree, E0, file=f)
f.close()

rpa0 = RPACorrelation('CO.ralda.pbe_wfcs_O.gpw',
                      txt='CO.ralda_04_O_rpa.txt',
                      wcomm=world.size)

E0_i = rpa0.calculate(ecut=400)

f = paropen('CO.ralda_rpa_O.dat', 'w')
for ecut, E0 in zip(rpa0.ecut_i, E0_i):
    print(ecut * Hartree, E0, file=f)
f.close()



import numpy as np
from ase.build import mx2
from gpaw import GPAW, FermiDirac
from gpaw.nlopt.matrixel import make_nlodata
from gpaw.nlopt.shg import get_shg

# Make the structure and add the vaccum around the layer
atoms = mx2(formula='MoS2', a=3.16, thickness=3.17, vacuum=5.0)
atoms.center(vacuum=15, axis=2)

# GPAW parameters:
nk = 40
params_gs = dict(
    mode='lcao',
    symmetry={'point_group': False, 'time_reversal': True},
    nbands='nao',
    convergence={'bands': -10},
    parallel={'domain': 1},
    occupations=FermiDirac(width=0.05),
    kpts={'size': (nk, nk, 1), 'gamma': True},
    xc='PBE',
    txt='gs.txt')

# Ground state calculation:
gs_name = 'gs.gpw'
calc = GPAW(**params_gs)
atoms.calc = calc
atoms.get_potential_energy()
atoms.calc.write(gs_name, mode='all')

# Calculate momentum matrix:
mml_name = 'mml.npz'
make_nlodata(gs_name=gs_name, out_name=mml_name)

# Shift parameters:
eta = 0.05  # Broadening in eV
w_ls = np.linspace(0, 6, 500)  # in eV
pol = 'yyy'

# LG calculation
shg_name1 = 'shg_' + pol + '_lg.npy'
get_shg(
    freqs=w_ls, eta=eta, pol=pol, gauge='lg',
    out_name=shg_name1, mml_name=mml_name)

# VG calculation
shg_name2 = 'shg_' + pol + '_vg.npy'
get_shg(
    freqs=w_ls, eta=eta, pol=pol, gauge='vg',
    out_name=shg_name2, mml_name=mml_name)



from ase.io import write
from gpaw import restart

basename = 'CO'

# load binary file and get calculator
atoms, calc = restart(basename + '.gpw')

# loop over all wfs and write their cube files
nbands = calc.get_number_of_bands()
for band in range(nbands):
    wf = calc.get_pseudo_wave_function(band=band)
    fname = f'{basename}_{band}.cube'
    print('writing wf', band, 'to file', fname)
    write(fname, atoms, data=wf)



import numpy as np
from ase.build import molecule

from gpaw import GPAW
from gpaw.poisson import PoissonSolver
from gpaw.lcaotddft import LCAOTDDFT
from gpaw.lcaotddft.dipolemomentwriter import DipoleMomentWriter
from gpaw.lcaotddft.densitymatrix import DensityMatrix
from gpaw.lcaotddft.frequencydensitymatrix import FrequencyDensityMatrix
from gpaw.lcaotddft.ksdecomposition import KohnShamDecomposition
from gpaw.tddft.folding import frequencies
from gpaw.tddft.units import au_to_eV
from gpaw.tddft.spectrum import photoabsorption_spectrum
from gpaw.mpi import world
from gpaw.test import equal


def test_lcaotddft_ksdecomp(in_tmp_dir):
    def relative_equal(x, y, *args, **kwargs):
        return equal(np.zeros_like(x), (x - y) / x, *args, **kwargs)

    # Atoms
    atoms = molecule('NaCl')
    atoms.center(vacuum=4.0)

    # Ground-state calculation
    calc = GPAW(nbands=6,
                h=0.4,
                setups=dict(Na='1'),
                basis='dzp',
                mode='lcao',
                poissonsolver=PoissonSolver(eps=1e-16),
                convergence={'density': 1e-8},
                txt='gs.out')
    atoms.calc = calc
    atoms.get_potential_energy()
    calc.write('gs.gpw', mode='all')

    # Time-propagation calculation
    td_calc = LCAOTDDFT('gs.gpw', txt='td.out')
    dmat = DensityMatrix(td_calc)
    freqs = frequencies(range(0, 31, 5), 'Gauss', 0.1)
    fdm = FrequencyDensityMatrix(td_calc, dmat, frequencies=freqs)
    DipoleMomentWriter(td_calc, 'dm.dat')
    kick_v = np.ones(3) * 1e-5
    td_calc.absorption_kick(kick_v)
    td_calc.propagate(20, 3)
    fdm.write('fdm.ulm')

    # Calculate reference spectrum
    photoabsorption_spectrum('dm.dat', 'spec.dat', delta_e=5, width=0.1)
    world.barrier()
    ref_wv = np.loadtxt('spec.dat')[:, 1:]

    # Calculate ground state with full unoccupied space
    calc = GPAW('gs.gpw', txt=None).fixed_density(
        nbands='nao', txt='unocc.out')
    calc.write('unocc.gpw', mode='all')

    # Construct KS electron-hole basis
    ksd = KohnShamDecomposition(calc)
    ksd.initialize(calc)
    ksd.write('ksd.ulm')

    # Load the objects
    calc = GPAW('unocc.gpw', txt=None)
    calc.initialize_positions()  # Initialize in order to calculate density
    ksd = KohnShamDecomposition(calc, 'ksd.ulm')
    dmat = DensityMatrix(calc)
    fdm = FrequencyDensityMatrix(calc, dmat, 'fdm.ulm')

    for w in range(1, len(fdm.freq_w)):
        rho_uMM = fdm.FReDrho_wuMM[w]
        rho_uMM = [rho_uMM[0] * kick_v[0] / np.sum(kick_v**2)]
        freq = fdm.freq_w[w]

        # Calculate dipole moment from density matrix
        rho_g = dmat.get_density([rho_uMM[0].imag])
        dm_v = dmat.density.finegd.calculate_dipole_moment(rho_g)
        spec_v = 2 * freq.freq / np.pi * dm_v / au_to_eV

        tol = 2e-3
        relative_equal(ref_wv[w], spec_v, tol)

        # KS transformation
        rho_up = ksd.transform(rho_uMM, broadcast=True)

        # Calculate dipole moment from induced density
        rho_g = ksd.get_density(calc.wfs, [rho_up[0].imag])
        dm_v = ksd.density.finegd.calculate_dipole_moment(rho_g)
        spec_v = 2 * freq.freq / np.pi * dm_v / au_to_eV

        tol = 2e-3
        relative_equal(ref_wv[w], spec_v, tol)

        # Calculate dipole moment from matrix elements
        dmrho_vp = ksd.get_dipole_moment_contributions(rho_up)
        spec_vp = 2 * freq.freq / np.pi * dmrho_vp.imag / au_to_eV
        spec_v = np.sum(spec_vp, axis=1)

        tol = 2e-3
        relative_equal(ref_wv[w], spec_v, tol)



# Creates: WS2_bands.png
import numpy as np
import matplotlib.pyplot as plt
from gpaw import GPAW
from gpaw.spinorbit import soc_eigenstates


calc = GPAW('WS2_bands.gpw', txt=None)
ef = calc.get_fermi_level()

x = np.loadtxt('WS2_kpath.dat')
X = np.loadtxt('WS2_highsym.dat')
e_kn = np.array([calc.get_eigenvalues(kpt=k)
                 for k in range(len(calc.get_ibz_k_points()))])
e_nk = e_kn.T
e_nk -= ef
for e_k in e_nk:
    plt.plot(x, e_k, '--', c='0.5')

soc = soc_eigenstates(calc)
e_kn = soc.eigenvalues()
s_knv = soc.spin_projections()
e_kn -= ef
s_nk = (s_knv[:, :, 2].T + 1.0) / 2.0

plt.xticks(X, [r'$\mathrm{M}$', r'$\mathrm{K}$', r'$\Gamma$',
               r'$\mathrm{-K}$', r'$\mathrm{-M}$'], size=20)
plt.yticks(size=20)
for i in range(len(X))[1:-1]:
    plt.plot(2 * [X[i]], [1.1 * np.min(e_kn), 1.1 * np.max(e_kn)],
             c='0.5', linewidth=0.5)

plt.scatter(np.tile(x, len(e_kn.T)), e_kn.T.reshape(-1),
            c=s_nk.reshape(-1),
            edgecolor=plt.get_cmap('jet')(s_nk.reshape(-1)),
            s=2,
            marker='+')

plt.ylabel(r'$\varepsilon_n(k)$ [eV]', size=24)
plt.axis([0, x[-1], -4.5, 4.5])
# plt.show()
plt.savefig('WS2_bands.png')



class ASECalculator:
    def __init__(self):
        self.atoms = None
        
    def get_potential_energy(self, atoms):
        if self.calculation_required(atoms, 'energy'):
            self.calculate(atoms)
            self.atoms = atoms.copy()  # store copy of last configuration
        return None

    ...



# Creates: qsfdtd_vs_mie.png
import numpy as np
import pylab as plt
from ase.units import Hartree, Bohr
from gpaw.fdtd.polarizable_material import PermittivityPlus, _eps0_au

# Nanosphere radius (Angstroms)
radius = 50.0

# Permittivity of Gold
# J. Chem. Phys. 135, 084121 (2011); http://dx.doi.org/10.1063/1.3626549
gold = [[0.2350, 0.1551, 95.62],
        [0.4411, 0.1480, -12.55],
        [0.7603, 1.946, -40.89],
        [1.161, 1.396, 17.22],
        [2.946, 1.183, 15.76],
        [4.161, 1.964, 36.63],
        [5.747, 1.958, 22.55],
        [7.912, 1.361, 81.04]]

# Plot calculated spectrum and compare with Mie theory
spec = np.loadtxt('spec.dat')
perm = PermittivityPlus(data=gold).value(spec[:, 0] / Hartree)
plt.figure()
plt.plot(spec[:, 0], spec[:, 1], 'r', label='QSFDTD')
plt.plot(spec[:, 0],
         3. * (4. / 3. * np.pi * (radius / Bohr)**3) *
         (spec[:, 0] / Hartree) / (2. * np.pi**2) / Hartree *
         np.imag((perm - _eps0_au) / (perm + 2. * _eps0_au)),
         'b', label='Mie theory')
plt.legend(loc=2)
plt.xlabel('Energy [eV]', fontsize=12)
plt.ylabel('Dipole strength [1/eV]', fontsize=12)
plt.xlim((0, 5.0))
plt.ylim((-1, 3500))
plt.savefig('qsfdtd_vs_mie.png')



import sys
from time import ctime
from pathlib import Path
import pickle

import numpy as np

from ase.units import Hartree
from ase.utils import convert_string_to_fd
from ase.utils.timing import Timer, timer

import gpaw.mpi as mpi
from gpaw.blacs import BlacsGrid, BlacsDescriptor, Redistributor
from gpaw.response.kspair import get_calc
from gpaw.response.kslrf import FrequencyDescriptor
from gpaw.response.chiks import ChiKS
from gpaw.response.kxc import get_fxc
from gpaw.response.kernels import get_coulomb_kernel


class FourComponentSusceptibilityTensor:
    """Class calculating the full four-component susceptibility tensor"""

    def __init__(self, gs, fxc='ALDA', fxckwargs={},
                 eta=0.2, ecut=50, gammacentered=False,
                 disable_point_group=True, disable_time_reversal=True,
                 bandsummation='pairwise', nbands=None,
                 bundle_integrals=True, bundle_kptpairs=False,
                 world=mpi.world, nblocks=1, txt=sys.stdout):
        """
        Currently, everything is in plane wave mode.
        If additional modes are implemented, maybe look to fxc to see how
        multiple modes can be supported.

        Parameters
        ----------
        gs : see gpaw.response.chiks, gpaw.response.kslrf
        fxc, fxckwargs : see gpaw.response.fxc
        eta, ecut, gammacentered
        disable_point_group,
        disable_time_reversal,
        bandsummation, nbands,
        bundle_integrals, bundle_kptpairs,
        world, nblocks, txt : see gpaw.response.chiks, gpaw.response.kslrf
        """
        # Initiate output file and timer
        self.world = world
        self.fd = convert_string_to_fd(txt, world)
        self.cfd = self.fd
        self.timer = Timer()

        # Load ground state calculation
        self.calc = get_calc(gs, fd=self.fd, timer=self.timer)

        # The plane wave basis is defined by keywords
        self.ecut = None if ecut is None else ecut / Hartree
        self.gammacentered = gammacentered

        # Initiate Kohn-Sham susceptibility and fxc objects
        self.chiks = ChiKS(self.calc, eta=eta, ecut=ecut,
                           gammacentered=gammacentered,
                           disable_point_group=disable_point_group,
                           disable_time_reversal=disable_time_reversal,
                           bandsummation=bandsummation, nbands=nbands,
                           bundle_integrals=bundle_integrals,
                           bundle_kptpairs=bundle_kptpairs, world=world,
                           nblocks=nblocks, txt=self.fd, timer=self.timer)
        self.fxc = get_fxc(self.calc, fxc,
                           response='susceptibility', mode='pw',
                           world=self.chiks.world, txt=self.chiks.fd,
                           timer=self.timer, **fxckwargs)

        # Parallelization over frequencies depends on the frequency input
        self.mynw = None
        self.w1 = None
        self.w2 = None

    def get_macroscopic_component(self, spincomponent, q_c, frequencies,
                                  filename=None, txt=None):
        """Calculates the spatially averaged (macroscopic) component of the
        susceptibility tensor and writes it to a file.

        Parameters
        ----------
        spincomponent, q_c,
        frequencies : see gpaw.response.chiks, gpaw.response.kslrf
        filename : str
            Save chiks_w and chi_w to file of given name.
            Defaults to:
            'chi%s_q«%+d-%+d-%+d».csv' % (spincomponent,
                                          *tuple((q_c * kd.N_c).round()))
        txt : str
            Save output of the calculation of this specific component into
            a file with the filename of the given input.

        Returns
        -------
        see calculate_macroscopic_component
        """

        if filename is None:
            tup = (spincomponent,
                   *tuple((q_c * self.calc.wfs.kd.N_c).round()))
            filename = 'chi%s_q«%+d-%+d-%+d».csv' % tup

        (omega_w,
         chiks_w,
         chi_w) = self.calculate_macroscopic_component(spincomponent, q_c,
                                                       frequencies,
                                                       txt=txt)

        write_macroscopic_component(omega_w, chiks_w, chi_w,
                                    filename, self.world)

        return omega_w, chiks_w, chi_w

    def calculate_macroscopic_component(self, spincomponent,
                                        q_c, frequencies, txt=None):
        """Calculates the spatially averaged (macroscopic) component of the
        susceptibility tensor.

        Parameters
        ----------
        spincomponent, q_c,
        frequencies : see gpaw.response.chiks, gpaw.response.kslrf
        txt : see get_macroscopic_component

        Returns
        -------
        omega_w, chiks_w, chi_w : nd.array, nd.array, nd.array
            omega_w: frequencies in eV
            chiks_w: macroscopic dynamic susceptibility (Kohn-Sham system)
            chi_w: macroscopic dynamic susceptibility
        """
        (pd, wd,
         chiks_wGG, chi_wGG) = self.calculate_component(spincomponent, q_c,
                                                        frequencies, txt=txt)

        # Macroscopic component
        chiks_w = chiks_wGG[:, 0, 0]
        chi_w = chi_wGG[:, 0, 0]

        # Collect data for all frequencies
        omega_w = wd.get_data() * Hartree
        chiks_w = self.collect(chiks_w)
        chi_w = self.collect(chi_w)

        return omega_w, chiks_w, chi_w

    def get_component_array(self, spincomponent, q_c, frequencies,
                            array_ecut=50, filename=None, txt=None):
        """Calculates a specific spin component of the susceptibility tensor,
        collects it as a numpy array in a reduced plane wave description
        and writes it to a file.

        Parameters
        ----------
        spincomponent, q_c,
        frequencies : see gpaw.response.chiks, gpaw.response.kslrf
        array_ecut : see calculate_component_array
        filename : str
            Save chiks_w and chi_w to pickle file of given name.
            Defaults to:
            'chi%sGG_q«%+d-%+d-%+d».pckl' % (spincomponent,
                                             *tuple((q_c * kd.N_c).round()))
        txt : str
            Save output of the calculation of this specific component into
            a file with the filename of the given input.

        Returns
        -------
        see calculate_component_array
        """

        if filename is None:
            tup = (spincomponent,
                   *tuple((q_c * self.calc.wfs.kd.N_c).round()))
            filename = 'chi%sGG_q«%+d-%+d-%+d».pckl' % tup

        (omega_w, G_Gc, chiks_wGG,
         chi_wGG) = self.calculate_component_array(spincomponent,
                                                   q_c,
                                                   frequencies,
                                                   array_ecut=array_ecut,
                                                   txt=txt)

        # Write susceptibilities to a pickle file
        write_component(omega_w, G_Gc, chiks_wGG, chi_wGG,
                        filename, self.world)

        return omega_w, G_Gc, chiks_wGG, chi_wGG

    def calculate_component_array(self, spincomponent, q_c, frequencies,
                                  array_ecut=50, txt=None):
        """Calculates a specific spin component of the susceptibility tensor
        and collects it as a numpy array in a reduced plane wave description.

        Parameters
        ----------
        spincomponent, q_c,
        frequencies : see gpaw.response.chiks, gpaw.response.kslrf
        array_ecut : float
            Energy cutoff for the reduced plane wave representation.
            The susceptibility is returned in the reduced representation.

        Returns
        -------
        omega_w, G_Gc, chiks_wGG, chi_wGG : nd.array(s)
            omega_w: frequencies in eV
            G_Gc : plane wave repr. as coordinates on the reciprocal lattice
            chiks_wGG: dynamic susceptibility (Kohn-Sham system)
            chi_wGG: dynamic susceptibility
        """
        (pd, wd,
         chiks_wGG, chi_wGG) = self.calculate_component(spincomponent, q_c,
                                                        frequencies, txt=txt)

        # Get frequencies in eV
        omega_w = wd.get_data() * Hartree

        # Get susceptibility in a reduced plane wave representation
        mask_G = get_pw_reduction_map(pd, array_ecut)
        chiks_wGG = np.ascontiguousarray(chiks_wGG[:, mask_G, :][:, :, mask_G])
        chi_wGG = np.ascontiguousarray(chi_wGG[:, mask_G, :][:, :, mask_G])

        # Get reduced plane wave repr. as coordinates on the reciprocal lattice
        G_Gc = get_pw_coordinates(pd)[mask_G]

        # Gather susceptibilities for all frequencies
        chiks_wGG = self.gather(chiks_wGG, wd)
        chi_wGG = self.gather(chi_wGG, wd)

        return omega_w, G_Gc, chiks_wGG, chi_wGG

    def calculate_component(self, spincomponent, q_c, frequencies, txt=None):
        """Calculate a single component of the susceptibility tensor.

        Parameters
        ----------
        spincomponent, q_c,
        frequencies : see gpaw.response.chiks, gpaw.response.kslrf

        Returns
        -------
        pd : PWDescriptor
            Descriptor object for the plane wave basis
        wd : FrequencyDescriptor
            Descriptor object for the calculated frequencies
        chiks_wGG : ndarray
            The process' block of the Kohn-Sham susceptibility component
        chi_wGG : ndarray
            The process' block of the full susceptibility component
        """
        pd = self.get_PWDescriptor(q_c)

        # Initiate new call-output file, if supplied
        if txt is not None:
            # Store timer and close old call-output file
            self.write_timer()
            if str(self.fd) != str(self.cfd):
                print('\nClosing, %s' % ctime(), file=self.cfd)
                self.cfd.close()
            # Initiate new output file
            self.cfd = convert_string_to_fd(txt, self.world)
        # Print to output file
        if str(self.fd) != str(self.cfd) or txt is not None:
            print('---------------', file=self.fd)
            print(f'Calculating susceptibility spincomponent={spincomponent}'
                  f'with q_c={pd.kd.bzk_kc[0]}', file=self.fd)

        wd = FrequencyDescriptor(np.asarray(frequencies) / Hartree)

        # Initialize parallelization over frequencies
        nw = len(wd)
        self.mynw = (nw + self.world.size - 1) // self.world.size
        self.w1 = min(self.mynw * self.world.rank, nw)
        self.w2 = min(self.w1 + self.mynw, nw)

        return self._calculate_component(spincomponent, pd, wd)

    def get_PWDescriptor(self, q_c):
        """Get the planewave descriptor defining the plane wave basis for the
        given momentum transfer q_c."""
        from gpaw.kpt_descriptor import KPointDescriptor
        from gpaw.wavefunctions.pw import PWDescriptor
        q_c = np.asarray(q_c, dtype=float)
        qd = KPointDescriptor([q_c])
        pd = PWDescriptor(self.ecut, self.calc.wfs.gd,
                          complex, qd, gammacentered=self.gammacentered)
        return pd

    def _calculate_component(self, spincomponent, pd, wd):
        """In-place calculation of the given spin-component."""
        if spincomponent in ['+-', '-+']:
            # No Hartree kernel
            K_GG = self.fxc(spincomponent, pd, txt=self.cfd)
        else:
            # Calculate Hartree kernel
            Kbare_G = get_coulomb_kernel(pd, self.calc.wfs.kd.N_c)
            vsqrt_G = Kbare_G ** 0.5
            K_GG = np.eye(len(vsqrt_G)) * vsqrt_G * vsqrt_G[:, np.newaxis]

            # Calculate exchange-correlation kernel
            Kxc_GG = self.fxc(spincomponent, pd, txt=self.cfd)
            if Kxc_GG is not None:
                K_GG += Kxc_GG
            
        chiks_wGG = self.calculate_ks_component(spincomponent, pd,
                                                wd, txt=self.cfd)

        chi_wGG = self.invert_dyson(chiks_wGG, K_GG)

        return pd, wd, chiks_wGG, chi_wGG

    def write_timer(self):
        """Write timer to call-output file and initiate a new."""
        self.timer.write(self.cfd)
        self.timer = Timer()

        # Update all other class instance timers
        self.chiks.timer = self.timer
        self.chiks.integrator.timer = self.timer
        self.chiks.kspair.timer = self.timer
        self.chiks.pme.timer = self.timer
        self.fxc.timer = self.timer

    def calculate_ks_component(self, spincomponent, pd, wd, txt=None):
        """Calculate a single component of the Kohn-Sham susceptibility tensor.

        Parameters
        ----------
        spincomponent : see gpaw.response.chiks, gpaw.response.kslrf
        pd, wd : see calculate_component

        Returns
        -------
        chiks_wGG : ndarray
            The process' block of the Kohn-Sham susceptibility component
        """
        # ChiKS calculates the susceptibility distributed over plane waves
        _, chiks_wGG = self.chiks.calculate(pd, wd, txt=txt,
                                            spincomponent=spincomponent)

        # Redistribute memory, so each block has its own frequencies, but all
        # plane waves (for easy invertion of the Dyson-like equation)
        chiks_wGG = self.distribute_frequencies(chiks_wGG)

        return chiks_wGG

    @timer('Invert dyson-like equation')
    def invert_dyson(self, chiks_wGG, Khxc_GG):
        """Invert the Dyson-like equation:

        chi = chi_ks + chi_ks Khxc chi
        """
        chi_wGG = np.empty_like(chiks_wGG)
        for w, chiks_GG in enumerate(chiks_wGG):
            chi_GG = np.dot(np.linalg.inv(np.eye(len(chiks_GG)) +
                                          np.dot(chiks_GG, Khxc_GG)),
                            chiks_GG)

            chi_wGG[w] = chi_GG

        return chi_wGG

    def collect(self, a_w):
        """Collect frequencies from all blocks"""
        # More documentation is needed! XXX
        world = self.chiks.world
        b_w = np.zeros(self.mynw, a_w.dtype)
        b_w[:self.w2 - self.w1] = a_w
        nw = len(self.chiks.omega_w)
        A_w = np.empty(world.size * self.mynw, a_w.dtype)
        world.all_gather(b_w, A_w)
        return A_w[:nw]

    def gather(self, A_wGG, wd):
        """Gather a full susceptibility array to root."""
        # Allocate arrays to gather (all need to be the same shape)
        shape = (self.mynw,) + A_wGG.shape[1:]
        tmp_wGG = np.empty(shape, dtype=A_wGG.dtype)
        tmp_wGG[:self.w2 - self.w1] = A_wGG

        # Allocate array for the gathered data
        if self.world.rank == 0:
            # Make room for all frequencies
            shape = (self.mynw * self.world.size,) + A_wGG.shape[1:]
            allA_wGG = np.empty(shape, dtype=A_wGG.dtype)
        else:
            allA_wGG = None

        self.world.gather(tmp_wGG, 0, allA_wGG)

        # Return array for w indeces on frequency grid
        if allA_wGG is not None:
            allA_wGG = allA_wGG[:len(wd), :, :]

        return allA_wGG

    @timer('Distribute frequencies')
    def distribute_frequencies(self, chiks_wGG):
        """Distribute frequencies to all cores."""
        # More documentation is needed! XXX
        world = self.chiks.world
        comm = self.chiks.interblockcomm

        if world.size == 1:
            return chiks_wGG

        nw = len(self.chiks.omega_w)
        nG = chiks_wGG.shape[2]
        mynw = (nw + world.size - 1) // world.size
        mynG = (nG + comm.size - 1) // comm.size

        wa = min(world.rank * mynw, nw)
        wb = min(wa + mynw, nw)

        if self.chiks.interblockcomm.size == 1:
            return chiks_wGG[wa:wb].copy()

        if self.chiks.intrablockcomm.rank == 0:
            bg1 = BlacsGrid(comm, 1, comm.size)
            in_wGG = chiks_wGG.reshape((nw, -1))
        else:
            bg1 = BlacsGrid(None, 1, 1)
            # bg1 = DryRunBlacsGrid(mpi.serial_comm, 1, 1)
            in_wGG = np.zeros((0, 0), complex)
        md1 = BlacsDescriptor(bg1, nw, nG**2, nw, mynG * nG)

        bg2 = BlacsGrid(world, world.size, 1)
        md2 = BlacsDescriptor(bg2, nw, nG**2, mynw, nG**2)

        r = Redistributor(world, md1, md2)
        shape = (wb - wa, nG, nG)
        out_wGG = np.empty(shape, complex)
        r.redistribute(in_wGG, out_wGG.reshape((wb - wa, nG**2)))

        return out_wGG

    def close(self):
        self.timer.write(self.cfd)
        print('\nClosing, %s' % ctime(), file=self.cfd)
        self.cfd.close()
        print('\nClosing, %s' % ctime(), file=self.fd)
        self.fd.close()


def get_pw_reduction_map(pd, ecut):
    """Get a mask to reduce the plane wave representation.

    Please remark, that the response code currently works with one q-vector
    at a time, at thus only a single plane wave representation at a time.

    Returns
    -------
    mask_G : nd.array (dtype=bool)
        Mask which reduces the representation
    """
    assert ecut is not None
    ecut /= Hartree
    assert ecut <= pd.ecut

    # List of all plane waves
    G_Gv = np.array([pd.G_Qv[Q] for Q in pd.Q_qG[0]])

    if pd.gammacentered:
        mask_G = ((G_Gv ** 2).sum(axis=1) <= 2 * ecut)
    else:
        mask_G = (((G_Gv + pd.K_qv[0]) ** 2).sum(axis=1) <= 2 * ecut)

    return mask_G


def get_pw_coordinates(pd):
    """Get the reciprocal lattice vector coordinates corresponding to a
    givne plane wave basis.

    Please remark, that the response code currently works with one q-vector
    at a time, at thus only a single plane wave representation at a time.

    Returns
    -------
    G_Gc : nd.array (dtype=int)
        Coordinates on the reciprocal lattice
    """
    # List of all plane waves
    G_Gv = np.array([pd.G_Qv[Q] for Q in pd.Q_qG[0]])

    # Use cell to get coordinates
    B_cv = 2.0 * np.pi * pd.gd.icell_cv
    return np.round(np.dot(G_Gv, np.linalg.inv(B_cv))).astype(int)


def write_macroscopic_component(omega_w, chiks_w, chi_w, filename, world):
    """Write the spatially averaged dynamic susceptibility."""
    assert isinstance(filename, str)
    if world.rank == 0:
        with Path(filename).open('w') as fd:
            for omega, chiks, chi in zip(omega_w, chiks_w, chi_w):
                print('%.6f, %.6f, %.6f, %.6f, %.6f' %
                      (omega, chiks.real, chiks.imag, chi.real, chi.imag),
                      file=fd)


def read_macroscopic_component(filename):
    """Read a stored macroscopic susceptibility file"""
    d = np.loadtxt(filename, delimiter=', ')
    omega_w = d[:, 0]
    chiks_w = np.array(d[:, 1], complex)
    chiks_w.imag = d[:, 2]
    chi_w = np.array(d[:, 3], complex)
    chi_w.imag = d[:, 4]

    return omega_w, chiks_w, chi_w


def write_component(omega_w, G_Gc, chiks_wGG, chi_wGG, filename, world):
    """Write the dynamic susceptibility as a pickle file."""
    assert isinstance(filename, str)
    if world.rank == 0:
        with open(filename, 'wb') as fd:
            pickle.dump((omega_w, G_Gc, chiks_wGG, chi_wGG), fd)


def read_component(filename):
    """Read a stored susceptibility component file"""
    assert isinstance(filename, str)
    with open(filename, 'rb') as fd:
        omega_w, G_Gc, chiks_wGG, chi_wGG = pickle.load(fd)

    return omega_w, G_Gc, chiks_wGG, chi_wGG



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Emacs: treat this as -*- python -*-
import numpy as np
import os
from optparse import OptionParser

parser = OptionParser(usage='%prog [options]', version='%prog 0.1')
parser.add_option('--dir', dest="dir", default='.', help='Results directory')
parser.add_option("--runs",
                  dest="runs",
                  default=5,
                  help='use that many runs to calculate the average.')
parser.add_option("--startcores",
                  dest="startcores",
                  default=1,
                  help='use at lease that many cores.')

opt, args = parser.parse_args()


colors = [
    'black', 'brown', 'red', 'orange', 'yellow', 'green', 'blue', 'violet',
    'gray', 'gray'
]


def plot(xdata, ydata, std, title, xlabel, ylabel, label, color, num=1):
    # matplotlib.use('Agg')
    import pylab

    # all goes to figure num
    pylab.figure(num=num, figsize=(7, 5.5))
    pylab.gca().set_position([0.10, 0.20, 0.85, 0.60])
    # let the plot have fixed y-axis scale
    miny = min(ydata)
    maxy = max(ydata)
    ywindow = maxy - miny
    pylab.gca().set_ylim(miny - ywindow / 4.0, maxy + ywindow / 3.0)
    pylab.bar(xdata, ydata, 0.3, yerr=std, label=label, color=color)
    pylab.title(title)
    pylab.xlabel(xlabel)
    pylab.ylabel(ylabel)


def plot_save(directory_name, out_prefix):
    from os.path import exists
    assert exists(directory_name)
    import pylab

    pylab.savefig(directory_name + os.path.sep + out_prefix + '.png')


def analyse_benchmark(ncores=8, startcores=1, machine='TEST', runs=7):
    system = machine + '_py'

    systems_string = {
        'carbon_py': 'gpaw 1865 on carbon',
        'carbon': 'mkl 10.0.2.018 dsyev on carbon',
        'niflheim_py': 'gpaw 1865 on niflheim',
        'niflheim': 'acml 4.0.1 dsyev on niflheim'
    }.get(system, False)

    processes = {
        'carbon_py': [1, 2, 4, 6, 8],
        'carbon': [1, 2, 4, 8],
        'niflheim_py': [1, 2, 3, 4],
        'niflheim': [1, 2, 4]
    }.get(system, False)

    if not systems_string:
        systems_string = 'gpaw on ' + machine
    if not processes:
        processes = [startcores]
        for n in range(startcores + 1, ncores + 1):
            if n % 2 == 0:
                processes.append(n)

    timer_entries_all = []
    if system.find('_py') == -1:
        for i in range(runs):
            timer_entries_all.append('run: ' + str(i))
    else:
        for i in range(runs):
            timer_entries_all.append('Run:  ' + str(i))

    import re

    # Select timer entries
    selected_entries = range(runs)

    height = {}

    gpaw_versions = []

    pre_results = {}
    results = {}

    timer_entries = []
    timer_entries_re = {}
    for entry in selected_entries:
        height[entry] = []
        timer_entries.append(timer_entries_all[entry])
        timer_entries_re[timer_entries_all[entry]] = re.compile(
            timer_entries_all[entry])

    # absolute path to directory
    root_abspath = os.path.abspath(opt.dir)
    # length of directory name

    ref_value_3300 = -44.85826
    ref_value_3301 = -44.85709
    # ref_value_3721 = -44.85666
    # ref_SCF_3721 = 35
    ref_value_5147 = -44.83504

    ref_value_6383 = -44.84197

    ref_value = ref_value_6383

    tolerance = 0.0001

    ref_failed = False
    h_failed = False
    for run in [str(p) + '_01' for p in processes]:
        # extract results
        rundir = os.path.join(root_abspath, system + run)
        file = os.path.join(rundir, 'out.txt')
        f = open(file, 'r')
        #
        print('Analysing ' + file, end=' ')
        #
        lines = f.readlines()
        # extract gpaw version
        for n, l in enumerate(lines):
            if l.startswith(' |__ |  _|___|_____|'):
                gpaw_version = lines[n + 0].strip().split()[3].split('.')[-1]
                break
        if gpaw_version[-1] == 'M':
            gpaw_version = gpaw_version[:-1]
        if gpaw_version.rfind(':') != -1:
            gpaw_version = gpaw_version[:gpaw_version.rfind(':')]
        gpaw_version = int(gpaw_version)
        if len(str(gpaw_version)) > 1:
            if gpaw_version <= 6383:
                ref_value = ref_value_6383
            if gpaw_version <= 5147:
                ref_value = ref_value_5147
            if gpaw_version <= 3720:
                ref_value = ref_value_3301
            if gpaw_version <= 3300:
                ref_value = ref_value_3300
        elif len(str(gpaw_version)) == 1:
            if gpaw_version <= 4:
                ref_value = ref_value_3300
                # ref_SCF = ref_SCF_3300
        gpaw_versions.append(gpaw_version)
        # search for timings
        print('gpaw version %d' % gpaw_version)
        for entry in selected_entries:
            h = []
            ref = []
            for line in lines:
                m = timer_entries_re[timer_entries_all[entry]].search(line)
                if m is not None:
                    h.append(float(line.split(':')[-1]))
                # break # stop after the first match
            for h_entry in h:
                if float(h_entry) < 0.0:
                    h_failed = True
                    break
            height[entry].append(h)
            for line in lines:
                m = re.compile('Zero').search(line)
                if m is not None:
                    ref.append(float(line.split(':')[-1]))
                # break # stop after the first match
            for ref_entry in ref:
                if abs(float(ref_entry) - ref_value) > tolerance:
                    ref_failed = True
                    break
    #
    assert len(processes) == len(gpaw_versions)
    for p in range(len(processes)):
        assert gpaw_versions[p] == max(
            gpaw_versions), 'incompatible gpaw versions across cores'
    #
    if h_failed:
        print('Panic: negative time in ' + file)
        assert not h_failed
    if ref_failed:
        print('Panic: wrong Zero Kelvin: value in ' + file + ' - should be ' +
              str(ref_value) + ' +- ' + str(tolerance))
        assert not ref_failed
    # arrange results
    for p in range(len(processes)):
        pre_results[processes[p]] = []
        for i in range(len(height)):
            pre_results[processes[p]].append(height[i][p])
    #
    # arrange results - calculate statistics
    for p in processes:
        results[p] = []
        temp = []
        for q in range(p):
            temp_q = []
            for i in range(len(pre_results[p])):
                # print pre_results[p][i][q]
                temp_q.append(pre_results[p][i][q])
                temp.append(pre_results[p][i][q])
            # averages for a given core q
            results[p].append((np.average(temp_q), np.std(temp_q)))
        # max, avrg, and std across all cores
        results[p].append(
            (np.average(temp), np.std(temp), min(temp), max(temp)))
    # for p in processes:
    #    #N = len(pre_results[p])
    #    #avg = sum(pre_results[p])/N
    #    #q = sqrt(sum([(x-avg)**2/(N) for x in pre_results[p]]))
    #    avg.append(np.average(pre_results[p]))
    #    q.append(np.std(pre_results[p]))
    import matplotlib
    matplotlib.use('Agg')
    from matplotlib import pylab
    # from http://matplotlib.sourceforge.net/examples/dashtick.py
    DASHBASE = 5
    DASHLEN = 25
    DASHSTAGGER = 3

    def dashlen(step):
        return DASHBASE + (DASHLEN * (step % DASHSTAGGER))

    # print scaling results
    parameters = processes
    zero = [0.0 for i in range(len(parameters))]
    pylab.plot(parameters, zero, 'k-', label='_nolegend_')
    ay1 = pylab.gca()
    ay1.xaxis.set_ticks(parameters)
    ay1.xaxis.set_ticklabels([str(x) for x in parameters])
    for p in processes:
        parameters = []
        avg = []
        std = []
        for i in range(len(results[p]) - 1):
            parameters.append(p + 0.3 * i)
            # avg and std across processes
            avg.append(results[p][i][0])
            std.append(results[p][i][1])
        # height
        # print parameters, avg, std
        print('No. of processes ' + str(int(parameters[0])) +
              ': time [sec]: avg ' + str(round(results[p][-1][0], 1)) +
              ', stddev ' + str(round(results[p][-1][1], 1)) + ', min ' +
              str(round(results[p][-1][2], 1)) + ', max ' +
              str(round(results[p][-1][3], 1)))
        plot(parameters,
             avg,
             std,
             systems_string + ' version ' + str(gpaw_version),
             'processes per node',
             'time [s]',
             'gpaw', (colors[p % 10]),
             num=1)
    # from two_scales.py
    plot_save(".", 'memory_bandwidth_' + system)
    pylab.close(1)


#

if __name__ == '__main__':
    from os import environ

    NCORES = int(environ.get('NCORES', 8))
    MACHINE = environ.get('MACHINE', 'TEST')
    assert NCORES >= 1, str(NCORES) + ' must be >= 1'

    runs = int(opt.runs)
    assert runs >= 1, runs + ' must be >= 1'
    startcores = int(opt.startcores)
    assert startcores >= 1, startcores + ' must be >= 1'

    analyse_benchmark(NCORES, startcores, MACHINE, runs=runs)



from ase.build import bulk
from gpaw import GPAW, PW, FermiDirac

calc = GPAW(mode=PW(600),
            xc='PBE',
            occupations=FermiDirac(0.01),
            spinpol=True,
            kpts=[8, 8, 8],
            parallel={'band': 1, 'domain': 1},
            txt='Fe_gs.txt')

bulk = bulk('Fe', 'bcc', a=2.87)
bulk.set_initial_magnetic_moments([1.0])
bulk.calc = calc
bulk.get_potential_energy()

calc.write('Fe_gs.gpw', mode='all')



from gpaw import GPAW, setup_paths
from gpaw.xas import XAS
import matplotlib.pyplot as plt

setup_paths.insert(0, '.')

dks_energy = 532.774  # from dks calcualtion

calc = GPAW('h2o_xas.gpw')

xas = XAS(calc, mode='xas')
x, y = xas.get_spectra(fwhm=0.5, linbroad=[4.5, -1.0, 5.0])
x_s, y_s = xas.get_spectra(stick=True)

shift = dks_energy - x_s[0]  # shift the first transition

y_tot = y[0] + y[1] + y[2]
y_tot_s = y_s[0] + y_s[1] + y_s[2]

plt.plot(x + shift, y_tot)
plt.bar(x_s + shift, y_tot_s, width=0.05)
plt.savefig('xas_h2o_spectrum.png')



from ase import Atoms
from gpaw import GPAW, FermiDirac, Davidson, Mixer
from gpaw.cdft.cdft import CDFT
from gpaw.cdft.cdft_coupling import CouplingParameters

# Set up the system
distance = 2.5
sys = Atoms('He2', positions=([0., 0., 0.], [0., 0., distance]))
sys.center(3)
sys.set_pbc(False)
sys.set_initial_magnetic_moments([0.5, 0.5])

# Calculator for the initial state
calc_a = GPAW(
    h=0.2,
    mode='fd',
    basis='dzp',
    charge=1,
    xc='PBE',
    symmetry='off',
    occupations=FermiDirac(0., fixmagmom=True),
    eigensolver=Davidson(3),
    spinpol=True,  # only spin-polarized calculations are supported
    nbands=4,
    mixer=Mixer(beta=0.25, nmaxold=3, weight=100.0),
    txt=f'He2+_initial_{distance:3.2f}.txt',
    convergence={
        'eigenstates': 1.0e-4,
        'density': 1.0e-1,
        'energy': 1e-1,
        'bands': 4})

# Set initial state cdft
cdft_a = CDFT(
    calc=calc_a,
    atoms=sys,
    charge_regions=[[0]],  # choose atom 0 as the constrained region
    charges=[1],  # constrain +1 charge
    charge_coefs=[2.7],  # initial guess for Vc
    method='L-BFGS-B',  # Vc optimization method
    txt=f'He2+_initial_{distance:3.2f}.cdft',  # cDFT output file
    minimizer_options={'gtol': 0.01})  # tolerance for cdft

# Get cdft energy
sys.calc = cdft_a
sys.get_potential_energy()

# the same for the final state
calc_b = GPAW(h=0.2,
              mode='fd',
              basis='dzp',
              charge=1,
              xc='PBE',
              symmetry='off',
              occupations=FermiDirac(0., fixmagmom=True),
              eigensolver=Davidson(3),
              spinpol=True,
              nbands=4,
              mixer=Mixer(beta=0.25, nmaxold=3, weight=100.0),
              txt=f'He2+_final_{distance:3.2f}.txt',
              convergence={
                  'eigenstates': 1.0e-4,
                  'density': 1.0e-1,
                  'energy': 1e-1,
                  'bands': 4})

cdft_b = CDFT(
    calc=calc_b,
    atoms=sys,
    charge_regions=[[1]],  # choose atom 1
    charges=[1],  # constrained charge +1
    charge_coefs=[2.7],
    method='L-BFGS-B',
    txt=f'He2+_final_{distance:3.2f}.cdft',
    minimizer_options={'gtol': 0.01})

sys.calc = cdft_b
sys.get_potential_energy()

# Now for the coupling parameter
coupling = CouplingParameters(cdft_a, cdft_b, AE=False)  # use pseudo orbitals
H12 = coupling.get_coupling_term()  # use original cDFT method



from gpaw.lcaotddft import LCAOTDDFT
from gpaw.lcaotddft.dipolemomentwriter import DipoleMomentWriter
from gpaw.lcaotddft.wfwriter import WaveFunctionWriter

# Read the restart file
td_calc = LCAOTDDFT('td.gpw', txt='tdc.out')

# Attach the data recording for appending the new data
DipoleMomentWriter(td_calc, 'dm.dat')
WaveFunctionWriter(td_calc, 'wf.ulm')

# Continue propagation
td_calc.propagate(20, 500)

# Save the state for restarting later
td_calc.write('td.gpw', mode='all')



from ase.build import bulk
from gpaw import GPAW
from gpaw.response.df import DielectricFunction

# Part 1: Ground state calculation
atoms = bulk('Al', 'fcc', a=4.043)  # generate fcc crystal structure
# GPAW calculator initialization:
k = 13
calc = GPAW(mode='pw', kpts=(k, k, k))

atoms.calc = calc
atoms.get_potential_energy()  # ground state calculation is performed
calc.write('Al.gpw', 'all')  # use 'all' option to write wavefunctions

# Part 2: Spectrum calculation
df = DielectricFunction(calc='Al.gpw')  # ground state gpw file as input

# Momentum transfer, must be the difference between two kpoints:
q_c = [1.0 / k, 0, 0]
df.get_eels_spectrum(q_c=q_c)  # by default, an 'eels.csv' file is generated



from gpaw.lcaotddft import LCAOTDDFT
from gpaw.lcaotddft.densitymatrix import DensityMatrix
from gpaw.lcaotddft.frequencydensitymatrix import FrequencyDensityMatrix
from gpaw.tddft.folding import frequencies

# Read the ground-state file
td_calc = LCAOTDDFT('gs.gpw')

# Attach analysis tools
dmat = DensityMatrix(td_calc)
freqs = frequencies([1.12, 2.48], 'Gauss', 0.1)
fdm = FrequencyDensityMatrix(td_calc, dmat, frequencies=freqs)

# Replay the propagation
td_calc.replay(name='wf.ulm', update='none')

# Store the density matrix
fdm.write('fdm.ulm')



"""Test hyperfine parameter CLI."""
import pytest
import numpy as np

from gpaw.hyperfine import main


@pytest.mark.parametrize('option', ['', '-x', '-d'])
@pytest.mark.serial
def test_hyperfine_cli(gpw_files, capsys, option):
    file = gpw_files['o2_pw']
    main(f'{file} -g O:-0.757 -u MHz {option}'.split())
    out = capsys.readouterr().out
    print(out)
    lines = [line.split() for line in out.splitlines()]
    if option == '-x':
        ref = [0.689, 16.87, -123.29, -123.29, 0.0, 0.0, 0.0]
    elif option == '-d':
        ref = [0.689, -39.00, -39.00, 101.15]
    else:
        ref = [0.689, 101.15, -39.00, -39.00, 0.0, 0.0, 0.0]
    for line in lines[3:5]:
        errors = (np.array([float(x) for x in line[2:]]) - ref)
        assert abs(errors).max() < 0.1
    assert float(lines[7][3]) == 2.0
    assert float(lines[10][1]) == -0.757



"""Bulk Al(fcc) test"""
from ase import Atoms
from ase.visualize import view
from gpaw import GPAW, PW

name = 'Al-fcc'
a = 4.05  # fcc lattice parameter
b = a / 2

bulk = Atoms('Al',
             cell=[[0, b, b],
                   [b, 0, b],
                   [b, b, 0]],
             pbc=True)

view(bulk)

k = 4
calc = GPAW(mode=PW(300),       # cutoff
            kpts=(k, k, k),     # k-points
            txt=name + '.txt')  # output file

bulk.calc = calc

energy = bulk.get_potential_energy()
calc.write(name + '.gpw')
print('Energy:', energy, 'eV')



"""Submit tests from the test suite that were removed because they were
too long."""

from myqueue.task import task


def create_tasks():
    return [
        task('H2Al110.py'),
        task('dscf_CO.py'),
        task('revtpss_tpss_scf.py'),
        task('ltt.py'),
        task('scalapack.py', cores=16),
        task('pblacs_oblong.py@64:5m')]



"""This script calculates the atomization energy of nitrogen using two
processes, each process working on a separate system."""
from gpaw import GPAW, mpi
import numpy as np
from ase import Atoms

cell = (8., 8., 8.)
p = 4.
separation = 1.103

rank = mpi.world.rank

# Master process calculates energy of N, while the other one takes N2
if rank == 0:
    system = Atoms('N', [(p, p, p)], magmoms=[3], cell=cell)
elif rank == 1:
    system = Atoms('N2', [(p, p, p + separation / 2.),
                          (p, p, p - separation / 2.)],
                   cell=cell)
else:
    raise Exception('This example uses only two processes')

# Open different files depending on rank
output = '%d.txt' % rank
calc = GPAW(communicator=[rank], txt=output, xc='PBE')
system.calc = calc
energy = system.get_potential_energy()

# Now send the energy from the second process to the first process,
if rank == 1:
    # Communicators work with arrays from Numeric only:
    mpi.world.send(np.array([energy]), 0)
else:
    # The first process receives the number and prints the atomization energy
    container = np.array([0.])
    mpi.world.receive(container, 1)

    # Ea = E[molecule] - 2 * E[atom]
    atomization_energy = container[0] - 2 * energy
    print(f'Atomization energy: {atomization_energy:.4f} eV')



import sys
import numpy as np
import matplotlib.pyplot as plt
for filename in sys.argv[1:]:
    a = np.loadtxt(filename, delimiter=',').T
    x = a[0]
    for y in a[1:]:
        plt.plot(x, y, '-')
plt.show()



# Copyright (C) 2006 CSC-Scientific Computing Ltd.

# Please see the accompanying LICENSE file for further information.
import os
import sys
import re
import distutils.util
from distutils.version import LooseVersion
from distutils.sysconfig import get_config_var, get_config_vars
from distutils.command.config import config
from glob import glob
from os.path import join
from stat import ST_MTIME

def check_packages(packages, msg, include_ase, import_numpy):
    """Check the python version and required extra packages

    If ASE is not installed, the `packages` list is extended with the
    ASE modules if they are found."""

    if sys.version_info < (2, 3, 0, 'final', 0):
        raise SystemExit('Python 2.3.1 or later is required!')

    if import_numpy:
        try:
            import numpy
        except ImportError:
            raise SystemExit('numpy is not installed!')
    else:
        msg += ['* numpy is not installed.',
                '  "include_dirs" in your customize.py must point to "numpy/core/include".']

    if not include_ase:
        if import_numpy:
            try:
                import ase
            except ImportError:
                import_ase = True
            else:
                import_ase = False
        else:
            import_ase = False

    if include_ase or import_ase:
        # Find ASE directories:
        # include_ase works in case:
        # cd gpaw # top-level gpaw source directory
        # tar zxf ~/python-ase-3.1.0.846.tar.gz
        # ln -s python-ase-3.1.0.846/ase .
        ase_root = 'ase'
        if include_ase:
            assert os.path.isdir(ase_root), ase_root+': No such file or directory'
        ase = []
        for root, dirs, files in os.walk(ase_root):
            if 'CVS' in dirs:
                dirs.remove('CVS')
            if '.svn' in dirs:
                dirs.remove('.svn')
            if '__init__.py' in files:
                ase.append(root.replace('/', '.'))

        if len(ase) == 0:
            msg += ['* ASE is not installed!  You may be able to install',
                    "  gpaw, but you can't use it without ASE!"]
        else:
            packages += ase

def find_file(arg, dir, files):
    #looks if the first element of the list arg is contained in the list files
    # and if so, appends dir to to arg. To be used with the os.path.walk
    if arg[0] in files:
        arg.append(dir)


def get_system_config(define_macros, undef_macros,
                      include_dirs, libraries, library_dirs, extra_link_args,
                      extra_compile_args, runtime_library_dirs, extra_objects,
                      msg, import_numpy):

    undef_macros += ['NDEBUG']
    if import_numpy:
        import numpy
        include_dirs += [numpy.get_include()]
    include_dirs += ['c/libxc']

    machine = os.uname()[4]
    if machine == 'sun4u':

        #  _
        # |_ | ||\ |
        #  _||_|| \|
        #

        extra_compile_args += ['-Kpic', '-fast']

        # Suppress warning from -fast (-xarch=native):
        f = open('cc-test.c', 'w')
        f.write('int main(){}\n')
        f.close()
        stderr = os.popen3('cc cc-test.c -fast')[2].read()
        arch = re.findall('-xarch=(\S+)', stderr)
        os.remove('cc-test.c')
        if len(arch) > 0:
            extra_compile_args += [f'-xarch={arch[-1]}']


        # We need the -Bstatic before the -lsunperf and -lfsu:
        # http://forum.java.sun.com/thread.jspa?threadID=5072537&messageID=9265782
        extra_link_args += ['-Bstatic', '-lsunperf', '-lfsu', '-Bdynamic']
        cc_version = os.popen3('cc -V')[2].readline().split()[3]
        if LooseVersion(cc_version) > '5.6':
            libraries.append('mtsk')
        else:
            extra_link_args.append('-lmtsk')
        #define_macros.append(('NO_C99_COMPLEX', '1'))

        msg += ['* Using SUN high performance library']

    elif sys.platform in ['aix5', 'aix6']:

        #
        # o|_  _ _
        # ||_)| | |
        #

        extra_compile_args += ['-qlanglvl=stdc99']
        # setting memory limit is necessary on aix5
        if sys.platform == 'aix5':
            extra_link_args += ['-bmaxdata:0x80000000',
                '-bmaxstack:0x80000000']

        libraries += ['f', 'lapack', 'essl']
        define_macros.append(('GPAW_AIX', '1'))

    elif machine == 'x86_64':

        #    _
        # \/|_||_    |_ |_|
        # /\|_||_| _ |_|  |
        #

        extra_compile_args += ['-Wall', '-std=c99']

        # Look for ACML libraries:
        acml = glob('/opt/acml*/g*64/lib')
        if len(acml) > 0:
            library_dirs += [acml[-1]]
            libraries += ['acml']
            if acml[-1].find('gfortran') != -1: libraries.append('gfortran')
            if acml[-1].find('gnu') != -1: libraries.append('g2c')
            extra_link_args += ['-Wl,-rpath=' + acml[-1]]
            msg += ['* Using ACML library']
        else:
            atlas = False
            for dir in ['/usr/lib', '/usr/local/lib']:
                if glob(join(dir, 'libatlas.a')) != []:
                    atlas = True
                    break
            if atlas:
                libraries += ['lapack', 'atlas', 'blas']
                library_dirs += [dir]
                msg +=  ['* Using ATLAS library']
            else:
                libraries += ['blas', 'lapack']
                msg +=  ['* Using standard lapack']

    elif machine =='ia64':

        #  _  _
        # |_ |  o
        #  _||_||
        #

        extra_compile_args += ['-Wall', '-std=c99']
        libraries += ['mkl','mkl_lapack64']

    elif machine == 'i686':

        #      _
        # o|_ |_||_
        # ||_||_||_|
        #

        extra_compile_args += ['-Wall', '-std=c99']

        if 'MKL_ROOT' in os.environ:
            mklbasedir = [os.environ['MKL_ROOT']]
        else:
            mklbasedir = glob('/opt/intel/mkl*')
        libs = ['libmkl_ia32.a']
        if mklbasedir != []:
            os.path.walk(mklbasedir[0],find_file, libs)
        libs.pop(0)
        if libs != []:
            libs.sort()
            libraries += ['mkl_lapack',
                          'mkl_ia32', 'guide', 'pthread', 'mkl']#, 'mkl_def']
            library_dirs += libs
            msg +=  [f'* Using MKL library: {library_dirs[-1]}']
            #extra_link_args += ['-Wl,-rpath=' + library_dirs[-1]]
        else:
            atlas = False
            for dir in ['/usr/lib', '/usr/local/lib']:
                if glob(join(dir, 'libatlas.a')) != []:
                    atlas = True
                    break
            if atlas:
                libraries += ['lapack', 'atlas', 'blas']
                library_dirs += [dir]
                msg +=  ['* Using ATLAS library']
            else:
                libraries += ['blas', 'lapack']
                msg +=  ['* Using standard lapack']

            # add libg2c if available
            g2c=False
            for dir in ['/usr/lib', '/usr/local/lib']:
                if glob(join(dir, 'libg2c.so')) != []:
                    g2c=True
                    break
                if glob(join(dir, 'libg2c.a')) != []:
                    g2c=True
                    break
            if g2c: libraries += ['g2c']

    elif sys.platform == 'darwin':

        extra_compile_args += ['-Wall', '-std=c99']
        include_dirs += ['/usr/include/malloc']

        if glob('/System/Library/Frameworks/vecLib.framework') != []:
            extra_link_args += ['-framework vecLib']
            msg += ['* Using vecLib']
        else:
            libraries += ['blas', 'lapack']
            msg +=  ['* Using standard lapack']

    return msg


def get_parallel_config(mpi_libraries,mpi_library_dirs,mpi_include_dirs,
                        mpi_runtime_library_dirs,mpi_define_macros):

    globals = {}
    exec(open('gpaw/mpi/config.py').read(), globals)
    mpi = globals['get_mpi_implementation']()

    if mpi == '':
        mpicompiler = None

    elif mpi == 'sun':
        mpi_include_dirs += ['/opt/SUNWhpc/include']
        mpi_libraries += ['mpi']
        mpi_library_dirs += ['/opt/SUNWhpc/lib']
        mpi_runtime_library_dirs += ['/opt/SUNWhpc/lib']
        mpicompiler = get_config_var('CC')

    elif mpi == 'poe':
        mpicompiler = 'mpcc_r'

    else:
        #Try to use mpicc
        mpicompiler = 'mpicc'

    return mpicompiler

def get_scalapack_config(define_macros):
    # check ScaLapack settings
    define_macros.append(('GPAW_WITH_SL', '1'))


def mtime(path, name, mtimes):
    """Return modification time.

    The modification time of a source file is returned.  If one of its
    dependencies is newer, the mtime of that file is returned.
    This function fails if two include files with the same name
    are present in different directories."""

    include = re.compile('^#\s*include "(\S+)"', re.MULTILINE)

    if name in mtimes:
        return mtimes[name]
    t = os.stat(os.path.join(path, name))[ST_MTIME]
    for name2 in include.findall(open(os.path.join(path, name)).read()):
        path2, name22 = os.path.split(name2)
        if name22 != name:
            t = max(t, mtime(os.path.join(path, path2), name22, mtimes))
    mtimes[name] = t
    return t

def check_dependencies(sources):
    # Distutils does not do deep dependencies correctly.  We take care of
    # that here so that "python setup.py build_ext" always does the right
    # thing!
    mtimes = {}  # modification times

    # Remove object files if any dependencies have changed:
    plat = distutils.util.get_platform() + '-' + sys.version[0:3]
    remove = False
    for source in sources:
        path, name = os.path.split(source)
        t = mtime(path + '/', name, mtimes)
        o = f'build/temp.{plat}/{source[:-2]}.o'  # object file
        if os.path.exists(o) and t > os.stat(o)[ST_MTIME]:
            print('removing', o)
            os.remove(o)
            remove = True

    so = f'build/lib.{plat}/_gpaw.so'
    if os.path.exists(so) and remove:
        # Remove shared object C-extension:
        # print 'removing', so
        os.remove(so)

def test_configuration():
    raise NotImplementedError


def write_configuration(define_macros, include_dirs, libraries, library_dirs,
                        extra_link_args, extra_compile_args,
                        runtime_library_dirs, extra_objects, mpicompiler,
                    mpi_libraries, mpi_library_dirs, mpi_include_dirs,
                    mpi_runtime_library_dirs, mpi_define_macros):

    # Write the compilation configuration into a file
    try:
        out = open('configuration.log', 'w')
    except IOError as x:
        print(x)
        return
    print("Current configuration", file=out)
    print("libraries", libraries, file=out)
    print("library_dirs", library_dirs, file=out)
    print("include_dirs", include_dirs, file=out)
    print("define_macros", define_macros, file=out)
    print("extra_link_args", extra_link_args, file=out)
    print("extra_compile_args", extra_compile_args, file=out)
    print("runtime_library_dirs", runtime_library_dirs, file=out)
    print("extra_objects", extra_objects, file=out)
    if mpicompiler is not None:
        print(file=out)
        print("Parallel configuration", file=out)
        print("mpicompiler", mpicompiler, file=out)
        print("mpi_libraries", mpi_libraries, file=out)
        print("mpi_library_dirs", mpi_library_dirs, file=out)
        print("mpi_include_dirs", mpi_include_dirs, file=out)
        print("mpi_define_macros", mpi_define_macros, file=out)
        print("mpi_runtime_library_dirs", mpi_runtime_library_dirs, file=out)
    out.close()


def build_interpreter(define_macros, include_dirs, libraries, library_dirs,
                      extra_link_args, extra_compile_args,
                      runtime_library_dirs, extra_objects,
                      mpicompiler, mpilinker, mpi_libraries, mpi_library_dirs,
                      mpi_include_dirs, mpi_runtime_library_dirs,
                      mpi_define_macros):

    #Build custom interpreter which is used for parallel calculations

    cfgDict = get_config_vars()
    plat = distutils.util.get_platform() + '-' + sys.version[0:3]

    cfiles = glob('c/[a-zA-Z_]*.c') + ['c/bmgs/bmgs.c']
    cfiles += glob('c/libxc/src/*.c')
    if ('HDF5', 1) in define_macros:
        cfiles += glob('h5py/c/*.c')
        cfiles += glob('h5py/c/lzf/*.c')

    cfiles2remove = ['c/libxc/src/test.c',
                     'c/libxc/src/xc_f.c',
                     'c/libxc/src/work_gga_x.c',
                     'c/libxc/src/work_lda.c'
                     ]

    for c2r in glob('c/libxc/src/funcs_*.c'): cfiles2remove.append(c2r)
    for c2r in cfiles2remove: cfiles.remove(c2r)
    sources = ['c/bc.c', 'c/localized_functions.c', 'c/mpi.c', 'c/_gpaw.c',
               'c/operators.c', 'c/transformers.c', 'c/compiled_WITH_SL.c',
               'c/blacs.c', 'c/utilities.c']
    objects = ' '.join(['build/temp.%s/' % plat + x[:-1] + 'o'
                        for x in cfiles])

    if not os.path.isdir(f'build/bin.{plat}/'):
        os.makedirs(f'build/bin.{plat}/')
    exefile = f'build/bin.{plat}/' + '/gpaw-python'

    # if you want static linked MPI libraries, uncomment the line below
    libraries += mpi_libraries
    library_dirs += mpi_library_dirs
    define_macros += mpi_define_macros
    include_dirs += mpi_include_dirs
    runtime_library_dirs += mpi_runtime_library_dirs

    define_macros.append(('PARALLEL', '1'))
    define_macros.append(('GPAW_INTERPRETER', '1'))
    macros = ' '.join(['-D%s=%s' % x for x in define_macros if x[0].strip()])

    include_dirs.append(cfgDict['INCLUDEPY'])
    include_dirs.append(cfgDict['CONFINCLUDEPY'])
    includes = ' '.join(['-I' + incdir for incdir in include_dirs])

    library_dirs.append(cfgDict['LIBPL'])
    lib_dirs = ' '.join(['-L' + lib for lib in library_dirs])

    libs = ' '.join(['-l' + lib for lib in libraries if lib.strip()])
    # BlueGene/P can statically link everything except
    # python, runtime and pthread library
    libs += ' -Wl,-dy'
    libpl = cfgDict['LIBPL']
    if glob(libpl + '/libpython*mpi*'):
        libs += f" -lpython{cfgDict['VERSION']}_mpi"
    else:
        libs += f" -lpython{cfgDict['VERSION']}"
    # libs += ' -lpython%s ' % cfgDict['VERSION']
    # if you want dynamic linked MPI libraries, uncomment the line below
    # only really needed for TAU profiling
    # libs += ' '.join(['-l' + lib for lib in mpi_libraries])
    libs += ' -lrt -lpthread'
    libs = ' '.join([libs, cfgDict['LIBS'], cfgDict['LIBM']])

    #Hack taken from distutils to determine option for runtime_libary_dirs
    if sys.platform[:6] == 'darwin':
        # MacOSX's linker doesn't understand the -R flag at all
        runtime_lib_option = '-L'
    elif sys.platform[:5] == 'hp-ux':
        runtime_lib_option = '+s -L'
    elif os.popen('mpicc --showme 2> /dev/null', 'r').read()[:3] == 'gcc':
        runtime_lib_option = '-Wl,-R'
    elif os.popen('mpicc -show 2> /dev/null', 'r').read()[:3] == 'gcc':
        runtime_lib_option = '-Wl,-R'
    else:
        runtime_lib_option = '-R'

    runtime_libs = ' '.join([ runtime_lib_option + lib for lib in runtime_library_dirs])

    extra_link_args.append(cfgDict['LDFLAGS'])
    if sys.platform in ['aix5', 'aix6']:
        extra_link_args.append(cfgDict['LINKFORSHARED'].replace('Modules', cfgDict['LIBPL']))
    elif sys.platform == 'darwin':
        pass
    else:
        extra_link_args.append(cfgDict['LINKFORSHARED'])

    if ('IO_WRAPPERS', 1) in define_macros:
        extra_link_args += ['-Wl,-wrap,fread',
                            '-Wl,-wrap,_IO_getc',
                            '-Wl,-wrap,getc_unlocked',
                            '-Wl,-wrap,fgets',
                            '-Wl,-wrap,ungetc',
                            '-Wl,-wrap,feof',
                            '-Wl,-wrap,ferror',
                            '-Wl,-wrap,fflush',
                            '-Wl,-wrap,fseek',
                            '-Wl,-wrap,rewind',
                            # '-Wl,-wrap,fileno',
                            '-Wl,-wrap,flockfile',
                            '-Wl,-wrap,funlockfile',
                            '-Wl,-wrap,clearerr',
                            '-Wl,-wrap,fgetpos',
                            '-Wl,-wrap,fsetpos',
                            '-Wl,-wrap,setbuf',
                            '-Wl,-wrap,setvbuf',
                            '-Wl,-wrap,ftell',
                            '-Wl,-wrap,fstat',
                            '-Wl,-wrap,fstat64',
                            '-Wl,-wrap,fgetc',
                            # '-Wl,-wrap,fputc',
                            # '-Wl,-wrap,fputs',
                            # '-Wl,-wrap,fwrite',
                            # '-Wl,-wrap,_IO_putc',
                            '-Wl,-wrap,fopen',
                            '-Wl,-wrap,fopen64',
                            '-Wl,-wrap,fclose',
                           ]

    # Compile the parallel sources
    for src in sources:
        obj = f'build/temp.{plat}/' + src[:-1] + 'o'
        cmd = ('%s %s %s %s -o %s -c %s ' ) % \
              (mpicompiler,
               macros,
               ' '.join(extra_compile_args),
               includes,
               obj,
               src)
        print(cmd)
        if '--dry-run' not in sys.argv:
            error=os.system(cmd)
            if error != 0:
                msg = ['* compiling FAILED!  Only serial version of code will work.']
                break

    # Link the custom interpreter
    cmd = ('%s -o %s %s %s %s %s %s %s' ) % \
          (mpilinker,
           exefile,
           objects,
           ' '.join(extra_objects),
           lib_dirs,
           libs,
           runtime_libs,
           ' '.join(extra_link_args))

    msg = ['* Building a custom interpreter']
    print(cmd)
    if '--dry-run' not in sys.argv:
        error=os.system(cmd)
        if error != 0:
            msg += ['* linking FAILED!  Only serial version of code will work.']


    return error, msg



"""Module for building atomic structures"""

from ase import Atom, Atoms


def bcc100(symbol, a, layers, L):
    """Build a bcc(100) surface

    symbol: chemical symbol ('H', 'Li', ...)
    a     : lattice constant
    layers: number of layers
    L     : height of unit cell"""

    a = float(a)

    # Distance between layers:
    z = a / 2

    assert L > layers * z, 'Unit cell too small!'

    # Start with an empty Atoms object with an orthorhombic unit cell:
    atoms = Atoms(pbc=(True, True, False), cell=(a, a, L))

    # Fill in the atoms:
    for n in range(layers):
        position = [a / 2 * (n % 2), a / 2 * (n % 2), n * z]
        atoms.append(Atom(symbol, position))

    atoms.center(axis=2)
    return atoms


if __name__ == '__main__':
    bcc = bcc100('Al', 4.0, 4, 15.0)
    from ase.visualize import view
    view(bcc * (4, 4, 2))



from ase.build import molecule
from gpaw import GPAW

a = 8.0
h = 0.2

energies = {}
resultfile = open(f'results-{h:.2f}.txt', 'w')

for name in ['H2O', 'H', 'O']:
    system = molecule(name)
    system.set_cell((a, a, a))
    system.center()

    calc = GPAW(h=h,
                txt=f'gpaw-{name}-{h:.2f}.txt')
    if name == 'H' or name == 'O':
        calc.set(hund=True)

    system.calc = calc

    energy = system.get_potential_energy()
    energies[name] = energy
    print(name, energy, file=resultfile)

e_atomization = energies['H2O'] - 2 * energies['H'] - energies['O']
print(e_atomization, file=resultfile)



from ase import Atoms
from ase.parallel import paropen
from gpaw import GPAW
from gpaw.mixer import Mixer
from gpaw.eigensolvers import CG
from gpaw.poisson import PoissonSolver
from gpaw import setup_paths
setup_paths.insert(0, '.')

# Usual GPAW definitions
h = 0.18
a = 12.00
c = a / 2

# XC functional + kinetic functional (minus the Tw contribution) to be used
xcname = '1.0_LDA_K_TF+1.0_LDA_X+1.0_LDA_C_PW'

# Fraction of Tw
lambda_coeff = 1.0

name = f'lambda_{lambda_coeff}'

filename = 'atoms_' + name + '.dat'

f = paropen(filename, 'w')

elements = ['N']

for symbol in elements:
    mixer = Mixer()

    eigensolver = CG(tw_coeff=lambda_coeff)

    poissonsolver = PoissonSolver()
    molecule = Atoms(symbol,
                     positions=[(c, c, c)],
                     cell=(a, a, a))

    calc = GPAW(h=h,
                xc=xcname,
                maxiter=240,
                eigensolver=eigensolver,
                mixer=mixer,
                setups=name,
                poissonsolver=poissonsolver)

    molecule.calc = calc

    E = molecule.get_total_energy()

    f.write(f'{symbol}\t{E}\n')



# Creates: zero.png, periodic.png, corrected.png, pwcorrected.png, slab.png
import numpy as np
import matplotlib.pyplot as plt
from ase.io import write
from gpaw import GPAW

# this test requires OpenEXR-libs

for name in ['zero', 'periodic', 'corrected', 'pwcorrected']:
    calc = GPAW(name + '.gpw', txt=None)

    efermi = calc.get_fermi_level()

    # Average over y and x:
    v = calc.get_electrostatic_potential().mean(1).mean(0)
    z = np.linspace(0, calc.atoms.cell[2, 2], len(v), endpoint=False)

    plt.figure(figsize=(6.5, 4.5))
    plt.plot(z, v, label='xy-averaged potential')
    plt.plot([0, z[-1]], [efermi, efermi], label='Fermi level')

    if name.endswith('corrected'):
        n = 6  # get the vacuum level 6 grid-points from the boundary
        plt.plot([0.2, 0.2], [efermi, v[n]], 'r:')
        plt.text(0.23, (efermi + v[n]) / 2,
                 r'$\phi$ = %.2f eV' % (v[n] - efermi), va='center')
        plt.plot([z[-1] - 0.2, z[-1] - 0.2], [efermi, v[-n]], 'r:')
        plt.text(z[-1] - 0.23, (efermi + v[-n]) / 2,
                 r'$\phi$ = %.2f eV' % (v[-n] - efermi),
                 va='center', ha='right')

    plt.xlabel(r'$z$, r$\AA$')
    plt.ylabel('(Pseudo) electrostatic potential, V')
    plt.xlim([0., z[-1]])
    if name == 'pwcorrected':
        title = 'PW-mode corrected'
    else:
        title = name.title()
    plt.title(title + ' boundary conditions')
    plt.savefig(name + '.png')

write('slab.pov',
      calc.atoms,
      rotation='-90x',
      show_unit_cell=2,
      povray_settings=dict(
          transparent=False,
          display=False)).render()



"""Make sure we get an exception when an atom is too close to the boundary."""
from ase import Atoms
from gpaw import GPAW
from gpaw.grid_descriptor import GridBoundsError


def test_pbc():
    a = 4.0
    x = 0.1
    hydrogen = Atoms('H', [(x, x, x)],
                     cell=(a, a, a),
                     calculator=GPAW(maxiter=7))
    try:
        hydrogen.get_potential_energy()
    except GridBoundsError:
        pass
    else:
        assert False



#!/usr/bin/env python3
"""GPAW-wrapper for qsub."""

import argparse
import subprocess

description = 'Submit a GPAW Python script via qsub.'
maxcores = 8  # XeonX5550


def main():
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('-p', '--processes', type=int, default=4,
                        help='Number of processes.')
    parser.add_argument('-t', '--time', type=int, default=1,
                        help='Max running time in hours.')
    parser.add_argument('-z', '--dry-run', action='store_true',
                        help='Don\'t actually submit script.')
    parser.add_argument('script', help='Python script')
    parser.add_argument('argument', nargs='*',
                        help='Command-line argument for Python script.')
    args = parser.parse_args()
    arguments = ' '.join(args.argument)
    cmd = f'gpaw python {args.script} {arguments}'
    nodes, rest = divmod(args.processes, maxcores)
    if nodes > 0:
        ppn = maxcores
        if rest > 0:
            msg = f'Please use a multiple of {maxcores} processes!'
            raise SystemExit(msg)
    else:
        nodes = 1
        ppn = args.processes
    lines = [
        '#!/bin/sh',
        '#PBS -q hpc',
        f'#PBS -N {args.script}',
        f'#PBS -l nodes={nodes}:ppn={ppn}',
        f'#PBS -l walltime={args.time}:00:00',
        # '#PBS -l feature=XeonX5550',
        'cd $PBS_O_WORKDIR',
        f'OMP_NUM_THREADS=1 mpiexec {cmd}']
    script = '\n'.join(lines) + '\n'
    if args.dry_run:
        print(script)
    else:
        p = subprocess.Popen(['qsub'], stdin=subprocess.PIPE)
        p.communicate(script.encode())


if __name__ == '__main__':
    main()



import numpy as np
from ase import Atoms
from gpaw import GPAW, PW, RMMDIIS
from gpaw.hybrids import HybridXC
from gpaw.mpi import world


def run(symbol, d0, M, ecut, L):
    a = Atoms(2 * symbol, cell=[L, L, L])
    a.positions[1, 0] = d0
    a.center()
    D = np.linspace(d0 * 0.98, d0 * 1.02, 5)
    E = []
    for d in D:
        a.set_distance(0, 1, d)
        a.calc = GPAW(
            mode=PW(ecut, force_complex_dtype=True),
            xc=HybridXC('PBE0'),
            nbands=0,
            # eigensolver='rmm-diis',
            txt=f'{symbol}2-{d / d0:.2f}-{ecut}-{L}.txt')
        e = a.get_potential_energy()
        E.append(e)

    p0 = np.polyfit(D, E, 3)
    p1 = np.polyder(p0)
    p2 = np.polyder(p1)
    d = sorted([(np.polyval(p2, d), d) for d in np.roots(p1)])[1][1]
    e2 = np.polyval(p0, d)

    a = Atoms(symbol, cell=[L, L, L], magmoms=[M])
    a.center()
    a.calc = GPAW(
        mode=PW(ecut, force_complex_dtype=True),
        xc=HybridXC('PBE0'),
        eigensolver=RMMDIIS(niter=1),
        parallel={'band': 1, 'kpt': 1},
        txt=f'{symbol}-{ecut}-{L}.txt')
    e1 = a.get_potential_energy()

    if world.rank == 0:
        print(symbol, ecut, L, 2 * e1 - e2, d)

    return 2 * e1 - e2, d


if __name__ == '__main__':
    for L in [8, 10, 12]:
        run('H', 0.75, 1, 500, L)
        run('N', 1.089, 3, 500, L)
    for ecut in [400, 500, 600, 700]:
        run('H', 0.75, 1, ecut, 8)
        run('N', 1.089, 3, ecut, 8)



from gpaw.tddft.spectrum import photoabsorption_spectrum
photoabsorption_spectrum('dm.dat', 'spec.dat',
                         folding='Gauss', width=0.1,
                         e_min=0.0, e_max=10.0, delta_e=0.01)



from sys import argv
import matplotlib.pyplot as plt

from ase.dft import STM
from gpaw import restart

filename = argv[1]
z0 = 8
bias = 1.0

atoms, calc = restart(filename, txt=None)

stm = STM(atoms, symmetries=[0, 1, 2])
c = stm.get_averaged_current(bias, z0)

print(f'Average current at z={z0:f}: {c:f}')

# Get 2d array of constant current heights:
x, y, h = stm.scan(bias, c)

print(f'Min: {h.min():.2f} Ang, Max: {h.max():.2f} Ang')

plt.contourf(x, y, h, 40)
plt.hot()
plt.colorbar()
plt.show()



from myqueue.task import task


def create_tasks():
    return [
        task('Be_gs_8bands.py@2:20m'),
        task('Be_8bands_lrtddft.py@2:20m', deps='Be_gs_8bands.py'),
        task('Be_8bands_lrtddft_dE.py@2:20m', deps='Be_gs_8bands.py'),
        task('Na2_relax_excited.py@4:8h')]



import pytest
from gpaw.utilities import compiled_with_sl
from ase.build import bulk
from ase.dft.kpoints import monkhorst_pack
from gpaw import GPAW, FermiDirac
from gpaw.test import equal
from gpaw.xc.fxc import FXCCorrelation
from gpaw.mpi import world, serial_comm

pytestmark = pytest.mark.skipif(
    world.size != 1 and not compiled_with_sl(),
    reason='world.size != 1 and not compiled_with_sl()')


def test_ralda_ralda_energy_Ni(in_tmp_dir):
    if world.rank == 0:
        Ni = bulk('Ni', 'fcc')
        Ni.set_initial_magnetic_moments([0.7])

        kpts = monkhorst_pack((3, 3, 3))

        calc = GPAW(mode='pw',
                    kpts=kpts,
                    occupations=FermiDirac(0.001),
                    setups={'Ni': '10'},
                    communicator=serial_comm)

        Ni.calc = calc
        Ni.get_potential_energy()
        calc.diagonalize_full_hamiltonian()
        calc.write('Ni.gpw', mode='all')

    world.barrier()

    rpa = FXCCorrelation('Ni.gpw', xc='RPA',
                         nfrequencies=8, skip_gamma=True)
    E_rpa = rpa.calculate(ecut=[50])

    ralda = FXCCorrelation('Ni.gpw', xc='rALDA', unit_cells=[2, 1, 1],
                           nfrequencies=8, skip_gamma=True)
    E_ralda = ralda.calculate(ecut=[50])

    rapbe = FXCCorrelation('Ni.gpw', xc='rAPBE', unit_cells=[2, 1, 1],
                           nfrequencies=8, skip_gamma=True)
    E_rapbe = rapbe.calculate(ecut=[50])

    equal(E_rpa, -7.827, 0.01)
    equal(E_ralda, -7.501, 0.01)
    equal(E_rapbe, -7.444, 0.01)



import os
from pathlib import Path

import pytest
from _pytest.tmpdir import _mk_tmp
from ase import Atoms
from ase.io import read
from gpaw.utilities import devnull
from ase.build import bulk

from gpaw import GPAW
from gpaw.cli.info import info
from gpaw.mpi import world, broadcast


@pytest.fixture
def in_tmp_dir(request, tmp_path_factory):
    """Run test in a temporary directory."""
    if world.rank == 0:
        path = _mk_tmp(request, tmp_path_factory)
    else:
        path = None
    path = broadcast(path)
    cwd = os.getcwd()
    os.chdir(path)
    try:
        yield path
    finally:
        os.chdir(cwd)


@pytest.fixture(scope='session')
def gpw_files(request, tmp_path_factory):
    """Reuse gpw-files.

    Returns a dict mapping names to paths to gpw-files.  If you
    want to reuse gpw-files from an earlier pytest session then set the
    ``$GPW_TEST_FILES`` environment variable and the files will be written
    to that folder.

    Example::

        def test_something(gpw_files):
            calc = GPAW(gpw_files['h2_lcao_wfs'])
            ...

    Possible systems are:

    * Bulk BCC-Li with 3x3x3 k-points: ``bcc_li_pw``, ``bcc_li_fd``,
      ``bcc_li_lcao``.

    * O2 molecule: ``o2_pw``.

    * H2 molecule: ``h2_pw``, ``h2_fd``, ``h2_lcao``.

    * H2 molecule (not centered): ``h2_pw_0``.

    * Spin-polarized H atom: ``h_pw``.

    * Polyethylene chain.  One unit, 3 k-points, no symmetry:
      ``c2h4_pw_nosym``.  Three units: ``c6h12_pw``.

    Files with wave functions are also availabel (add ``_wfs`` to the names).
    """
    path = os.environ.get('GPW_TEST_FILES')
    if path is None:
        if world.rank == 0:
            path = _mk_tmp(request, tmp_path_factory)
        else:
            path = None
        path = broadcast(path)
    return GPWFiles(Path(path))


class GPWFiles:
    """Create gpw-files."""
    def __init__(self, path: Path):
        self.path = path
        self.gpw_files = {}
        for file in path.glob('*.gpw'):
            self.gpw_files[file.name[:-4]] = file

    def __getitem__(self, name):
        if name not in self.gpw_files:
            rawname, _, _ = name.partition('_wfs')
            calc = getattr(self, rawname)()
            path = self.path / (rawname + '.gpw')
            calc.write(path)
            self.gpw_files[rawname] = path
            path = self.path / (rawname + '_wfs.gpw')
            calc.write(path, mode='all')
            self.gpw_files[rawname + '_wfs'] = path
        return self.gpw_files[name]

    def bcc_li_pw(self):
        return self.bcc_li({'name': 'pw', 'ecut': 200})

    def bcc_li_fd(self):
        return self.bcc_li({'name': 'fd'})

    def bcc_li_lcao(self):
        return self.bcc_li({'name': 'lcao'})

    def bcc_li(self, mode):
        li = bulk('Li', 'bcc', 3.49)
        li.calc = GPAW(mode=mode,
                       kpts=(3, 3, 3),
                       txt=self.path / f'bcc_li_{mode["name"]}.txt')
        li.get_potential_energy()
        return li.calc

    def h2_pw(self):
        return self.h2({'name': 'pw', 'ecut': 200})

    def h2_fd(self):
        return self.h2({'name': 'fd'})

    def h2_lcao(self):
        return self.h2({'name': 'lcao'})

    def h2(self, mode):
        h2 = Atoms('H2', positions=[[0, 0, 0], [0.74, 0, 0]])
        h2.center(vacuum=2.5)
        h2.calc = GPAW(mode=mode,
                       txt=self.path / f'h2_{mode["name"]}.txt')
        h2.get_potential_energy()
        return h2.calc

    def h2_pw_0(self):
        h2 = Atoms('H2',
                   positions=[[-0.37, 0, 0], [0.37, 0, 0]],
                   cell=[5.74, 5, 5])
        h2.calc = GPAW(mode={'name': 'pw', 'ecut': 200},
                       txt=self.path / 'h2_pw_0.txt')
        h2.get_potential_energy()
        return h2.calc

    def h_pw(self):
        h = Atoms('H', magmoms=[1])
        h.center(vacuum=4.0)
        h.calc = GPAW(mode={'name': 'pw', 'ecut': 500},
                      txt=self.path / 'h_pw.txt')
        h.get_potential_energy()
        return h.calc

    def o2_pw(self):
        d = 1.1
        h = Atoms('O2', positions=[[0, 0, 0], [d, 0, 0]], magmoms=[1, 1])
        h.center(vacuum=4.0)
        h.calc = GPAW(mode={'name': 'pw', 'ecut': 800},
                      txt=self.path / 'o2_pw.txt')
        h.get_potential_energy()
        return h.calc

    def c2h4_pw_nosym(self):
        d = 1.54
        h = 1.1
        x = d * (2 / 3)**0.5
        z = d / 3**0.5
        pe = Atoms('C2H4',
                   positions=[[0, 0, 0],
                              [x, 0, z],
                              [0, -h * (2 / 3)**0.5, -h / 3**0.5],
                              [0, h * (2 / 3)**0.5, -h / 3**0.5],
                              [x, -h * (2 / 3)**0.5, z + h / 3**0.5],
                              [x, h * (2 / 3)**0.5, z + h / 3**0.5]],
                   cell=[2 * x, 0, 0],
                   pbc=(1, 0, 0))
        pe.center(vacuum=2.0, axis=(1, 2))
        pe.calc = GPAW(mode='pw',
                       kpts=(3, 1, 1),
                       symmetry='off',
                       txt=self.path / 'c2h4_pw_nosym.txt')
        pe.get_potential_energy()
        return pe.calc

    def c6h12_pw(self):
        pe = read(self['c2h4_pw_nosym'])
        pe = pe.repeat((3, 1, 1))
        pe.calc = GPAW(mode='pw', txt=self.path / 'c6h12_pw.txt')
        pe.get_potential_energy()
        return pe.calc

    def h2o_lcao(self):
        from ase.build import molecule
        atoms = molecule('H2O', cell=[8, 8, 8], pbc=1)
        atoms.center()
        atoms.calc = GPAW(mode='lcao', txt='h2o.txt')
        atoms.get_potential_energy()
        return atoms.calc


class GPAWPlugin:
    def __init__(self):
        if world.rank == -1:
            print()
            info()

    def pytest_terminal_summary(self, terminalreporter, exitstatus, config):
        from gpaw.mpi import size
        terminalreporter.section('GPAW-MPI stuff')
        terminalreporter.write(f'size: {size}\n')


def pytest_configure(config):
    if world.rank != 0:
        try:
            tw = config.get_terminal_writer()
        except AttributeError:
            pass
        else:
            tw._file = devnull
    config.pluginmanager.register(GPAWPlugin(), 'pytest_gpaw')


def pytest_runtest_setup(item):
    """Skip some tests.

    If:

    * they depend on libxc and GPAW is not compiled with libxc
    * they are before $PYTEST_START_AFTER
    """
    from gpaw import libraries

    if world.size > 1:
        for mark in item.iter_markers():
            if mark.name == 'serial':
                pytest.skip('Only run in serial')

    if item.location[0] <= os.environ.get('PYTEST_START_AFTER', ''):
        pytest.skip('Not after $PYTEST_START_AFTER')
        return

    if libraries['libxc']:
        return

    if any(mark.name in {'libxc', 'mgga'}
           for mark in item.iter_markers()):
        pytest.skip('No LibXC.')



import pytest
import numpy as np

from ase.build import molecule
from gpaw import GPAW
from gpaw.lcaotddft import LCAOTDDFT
from gpaw.poisson import PoissonSolver
from gpaw.lcaotddft.dipolemomentwriter import DipoleMomentWriter
from gpaw.mpi import world
from gpaw.utilities import compiled_with_sl

from gpaw.test import equal

# Atoms
pytestmark = pytest.mark.skipif(
    world.size == 1 or not compiled_with_sl(),
    reason='world.size == 1 or not compiled_with_sl()')


def test_lcaotddft_parallel_options(in_tmp_dir):
    atoms = molecule('NaCl')
    atoms.center(vacuum=4.0)

    # Ground-state calculation
    calc = GPAW(nbands=7, h=0.4, setups=dict(Na='1'),
                basis='dzp', mode='lcao',
                poissonsolver=PoissonSolver(eps=1e-16),
                convergence={'density': 1e-8},
                txt='gs.out')
    atoms.calc = calc
    _ = atoms.get_potential_energy()
    calc.write('gs.gpw', mode='all')

    # Reference time-propagation calculation
    td_calc = LCAOTDDFT('gs.gpw', txt='td.out')
    DipoleMomentWriter(td_calc, 'dm.dat')
    td_calc.absorption_kick(np.ones(3) * 1e-5)
    td_calc.propagate(20, 3)

    # Check dipole moment file
    world.barrier()
    ref = np.loadtxt('dm.dat').ravel()

    # Test parallelization options
    par_i = []

    if world.size == 2:
        par_i.append({'band': 2})
        par_i.append({'sl_default': (2, 1, 2)})
        par_i.append({'sl_default': (1, 2, 4), 'band': 2})
    elif world.size == 4:
        par_i.append({'band': 2})
        par_i.append({'sl_default': (2, 2, 2)})
        par_i.append({'sl_default': (2, 2, 4), 'band': 2})
    else:
        par_i.append({'band': 2})
        par_i.append({'sl_auto': True})
        par_i.append({'sl_auto': True, 'band': 2})

    for i, par in enumerate(par_i):
        td_calc = LCAOTDDFT('gs.gpw', parallel=par, txt='td%d.out' % i)
        DipoleMomentWriter(td_calc, 'dm%d.dat' % i)
        td_calc.absorption_kick(np.ones(3) * 1e-5)
        td_calc.propagate(20, 3)

        world.barrier()
        data = np.loadtxt('dm%d.dat' % i).ravel()

        tol = 1e-11
        equal(data, ref, tol)



# %%
# teacher
import ase.visualize as viz
viz.view = lambda atoms, repeat=None: None

# %%
"""
# Non-collinear magnetism - VI$_2$

Having looked at the ferromagnetic compound CrI$_3$, we now move on to a bit
more complicated material. We will stay in the framework of 2D materials, but
now move on to anti-ferromagnetic exchange coupling. We will still have the
Hamiltonian

$$H = -\frac{1}{2}\sum_{ij}J_{ij}\mathbf{S}_i\cdot \mathbf{S}_j+A\sum_i(S_i^z)^2$$

in mind, but now with $J<0$. Go to the 2D database at
https://cmrdb.fysik.dtu.dk/?project=c2db and search for VI$_2$ in the CdI$_2$
prototype. Click on the *ferromagnetic* structure and download the .xyz file.
Since we will need to do LDA calculations later on, we start by relaxing the
structure with the LDA functional. We will be interested in the anti-
ferromagnetic state later, but perform the relaxation in the ferromagnetic
state, which has a smaller unit cell. Fill in the missing pieces and run the
cell below. V has the electronic configuration [Ar]3d$^3$4s$^2$, which can be
used to guess the initial magnetic moments. The calculation takes about 17
minutes.

"""

# %%
from ase.io import read
from ase.visualize import view
from ase.optimize import BFGS
from ase.constraints import UnitCellFilter
from gpaw import GPAW, PW

S = 3 / 2  # student: S = ???
m = S * 2
layer = read('VI2.xyz')
layer.set_initial_magnetic_moments([m, 0, 0])
view(layer)

calc = GPAW(mode=PW(400), xc='LDA', kpts=(4, 4, 1))
layer.calc = calc

uf = UnitCellFilter(layer, mask=[1, 1, 0, 0, 0, 1])
opt = BFGS(uf)
opt.run(fmax=0.1)

calc.set(symmetry='off')
calc.get_potential_energy()

calc.write('VI2_relaxed.gpw')

# %%
"""
### Magnetic anisotropy

Note that we switch off symmetry in the end of the script and do a last
calculations with all $k$-points in the Brillouin zone. This is because
spinors transform in a non-trivial way and the spin-orbit coupling can be
obtained from the irreducible $k$-points without transforming the
wavefunctions at symmetry related $k$-points. Evaluate the magnetic
anisotropy in the cell below in a manner similar to the case of CrI$_3$. Is
the easy-axis in plane or out of plane? Do you expect to find a finite
critical temperature based on this?
"""

# %%
from gpaw.spinorbit import soc_eigenstates
# teacher:

calc = GPAW('VI2_relaxed.gpw')
e_x, e_y, e_z = (
    soc_eigenstates(calc, theta=theta, phi=phi).calculate_band_energy()
    for theta, phi in [(90, 0), (90, 90), (0, 0)])
de_zx = e_z - e_x
de_zy = e_z - e_y
print(f'dE_zx = {de_zx * 1000:1.3f} meV')
print(f'dE_zy = {de_zy * 1000:1.3f} meV')
A = de_zx

# %%
"""
### Anti-ferromagnetic state

We can do an anti-ferromagnetic calculation by repeating the structure we
just relaxed and setting the initial magnetic moments accordingly. Fill in
the missing values in the cell below and run it. The calculation takes about
7 minutes.

"""

# %%
layer_afm = layer.repeat((2, 1, 1))
layer_afm.set_initial_magnetic_moments([m, 0, 0, -m, 0, 0])
view(layer_afm)

calc = GPAW(mode=PW(400),
            xc='LDA',
            kpts=(2, 4, 1))  # student: kpts=???))
layer_afm.calc = calc
layer_afm.get_potential_energy()
calc.write('V2I4_afm.gpw')

...
# teacher:
layer_fm = layer.repeat((2, 1, 1))
layer_fm.set_initial_magnetic_moments([m, 0, 0, m, 0, 0])
calc = GPAW(mode=PW(400), xc='LDA', kpts=(2, 4, 1))
layer_fm.calc = calc
layer_fm.get_potential_energy()
calc.write('V2I4_fm.gpw')

# %%
"""
### Calculating J

Is the total energy of anti-ferromagnetic state smaller than the
ferromagnetic one? It should be. But since we are running with rather low
parameters for $k$-point sampling and plane wave cutoff, we better perform a
ferromagnetic calculation with exactly the same parameters to make sure. Run
the cell above with in a ferromagnetic spin state and compare the resulting
energies.

The anti-ferromagnetic state we constructed appears to have a lower energy,
but can we really be sure that this is the magnetic ground state? The
exchange coupling must be negative, which indicates that spins prefer to be
anti-aligned. Draw the magnetic configuration of the lattice on a piece of
paper and convince yourself that all spins cannot be anti-aligned on the
hexagonal lattice. The anti-ferromagnetic structure we obtained must thus be
frustrated and possibly not the true ground state.

Let us put that aside for the moment and try to calculate $J$. Use the
Heisenberg model with classical spins, nearest neighbor interaction only, and
$A=0$ to derive that the energy per site of the two configurations can be
written as

$$E_{\mathrm{FM}} = E_0 - \frac{1}{2}6S^2J$$

and

$$E_{\mathrm{AFM}} = E_0 + \frac{1}{2}2S^2J$$

per site, where $E_0$ is some reference energy. Use these expressions to
eliminate $E_0$ and express $J$ in terms of the energy difference. Use the
energies obtained with DFT to calculate $J$. You should get -1.4 meV. Do it
with python in the cell below.
"""

# %%
E_fm = layer_fm.get_potential_energy() / 2  # student: E_fm = ???
E_afm = layer_afm.get_potential_energy() / 2  # student: E_afm = ???
dE = E_afm - E_fm  # student:
J = dE / 4 / S**2  # student: J = ???
print(f'J = {J * 1000:1.2f} meV')

# %%
"""
### Non-collinear configuration

As it turn out the optimal spin structure of a hexagonal lattice with anti-
ferromagntice coupling is taking all spins at 120$^\circ$ angles with respect
to each other. Draw this structure and convince yourself that it can be done.

1. What is the minimal number of magnetic atoms required in the magnetic unit cell
2. Verrify that the Heisenberg model with classical spins gives a lower energy with this configuration that the anti-aligned structure calculated above. The energy per site of this state should be

$$E_{\mathrm{NC}}=E_0+\frac{3}{2}S^2J.$$

We will now check if LDA can verify this prediction. To do that we need to
perform a calculation with non-collinear spin. This is done in the cell
below. Assert that the the total energy per site is lower than what we
obtained with the collinear anti-ferromagnetic configuration above. Also
check the local magnetic moments printet at the end of the calculation.
"""

# %%
import numpy as np
from ase.io import read
from ase.visualize import view
from gpaw import GPAW, PW, MixerDif

m = 3
cell_cv = layer.get_cell()
layer_nc = layer.repeat((3, 1, 1))
new_cell_cv = [[3 * cell_cv[0, 0] / 2, 3**0.5 * cell_cv[0, 0] / 2, 0.0],
               [3 * cell_cv[0, 0] / 2, -3**0.5 * cell_cv[0, 0] / 2, 0.0],
               [0.0, 0.0, cell_cv[2, 2]]]
layer_nc.set_cell(new_cell_cv)
view(layer_nc)

magmoms = np.zeros((len(layer_nc), 3), float)
magmoms[0] = [m, 0, 0]
magmoms[3] = [m * np.cos(2 * np.pi / 3), m * np.sin(2 * np.pi / 3), 0]
magmoms[6] = [m * np.cos(2 * np.pi / 3), -m * np.sin(2 * np.pi / 3), 0]

calc = GPAW(mode=PW(400),
            xc='LDA',
            mixer=MixerDif(),
            symmetry='off',
            experimental={'magmoms': magmoms, 'soc': False},
            parallel={'domain': 1, 'band': 1},
            kpts=(2, 2, 1))
layer_nc.calc = calc
layer_nc.get_potential_energy()
calc.write('nc_nosoc.gpw')

# teacher:
magmoms = np.zeros((len(layer_nc), 3), float)
magmoms[0] = [m, 0, 0]
magmoms[3] = [m, 0, 0]
magmoms[6] = [m, 0, 0]
calc = GPAW(mode=PW(400),
            xc='LDA',
            mixer=MixerDif(),
            symmetry='off',
            experimental={'magmoms': magmoms, 'soc': False},
            parallel={'domain': 1, 'band': 1},
            kpts=(2, 2, 1))
layer_nc.calc = calc
layer_nc.get_potential_energy()
calc.write('fm_nosoc.gpw')

# %%
"""
### Anisotropy and exhange coupling from the non-collinear configuration

In the cell above we could have set 'soc'=True' to include spin-orbit
coupling in the self-consistent non-collinear solution. However, it is more
convenient for us to exclude it such that we can explicitly obtain the
anisotropy based on this calculation.

If the Heisenberg Hamiltonian with nearest neighbor interactions is a good
model we should be able to obtain both $J$ and $A$ from the non-collinear
calculation as well. Write some python code in the cell below that return $J$
and $A$ based on the non-collinear calculation. The calculation of $J$
requires two spin configurations and we could use the ferromagnetic
calculation in the simple unit cell obtained at the top of the notebook as
one of them. But, since the energy differences are rather small it is much
better if you can obtain a ferromagnetic state with the same unit cell and
parameters as we used for the non-collinear calculation. You may thus run the
cell above once more, but with ferromagnetic alignment.
"""

# %%

# teacher:
calc = GPAW('nc_nosoc.gpw', txt=None)
E_nc = calc.get_potential_energy() / 3
e_x, e_y, e_z = (
    soc_eigenstates(calc, theta=theta, phi=phi).calculate_band_energy() / 3
    for theta, phi in [(90, 0), (90, 90), (0, 0)])
de_zx = e_z - e_x
de_zy = e_z - e_y
print(f'NC: A_zx = {de_zx * 1000:1.3f} meV')
print(f'NC: A_zy = {de_zy * 1000:1.3f} meV')
print()

calc = GPAW('fm_nosoc.gpw', txt=None)
E_fm = calc.get_potential_energy() / 3
e_x, e_y, e_z = (
    soc_eigenstates(calc, theta=theta, phi=phi).calculate_band_energy() / 3
    for theta, phi in [(90, 0), (90, 90), (0, 0)])
de_zx = e_z - e_x
de_zy = e_z - e_y
print(f'FM: A_zx = {de_zx * 1000:1.3f} meV')
print(f'FM: A_zy = {de_zy * 1000:1.3f} meV')
print()

dE = E_nc - E_fm
J = 2 * dE / 9 / S**2
print(f'J = {J * 1000:1.2f} meV')
print()
calc = GPAW('VI2_relaxed.gpw', txt=None)
E_fm = calc.get_potential_energy()
dE = E_nc - E_fm
J = 2 * dE / 9 / S**2
print(f'J = {J * 1000:1.2f} meV')


# %%
"""
### Critical temperature?
Now answer the following questions:

1. What is the easy axis? (Hint: the anisotropy is calculated by rotating
   the initial spin configuration first by $\theta$ and then by $\varphi$).
   Does it agree with waht you found above for the simple ferromagnetic state?

2. Is there any rotational freedom left in the non-collinear ground state if
   we assume in plane isotropy?

You might be able to convince yourself that some degree of in-plane
anisotropy is required as well to obtain a finite critical temperature for
magnetic order.

Clearly the non-collinear spin state of VI$_2$ is more difficult to describe
than the ferromagnetic state in CrI$_3$ and we do not yet have a simple
theoretical expression fot the critical temperature as a function of
anisotropy and exchange coupling constants. However, with the rapid
development of excperimental techniques to synthesize and characterize 2D
materials it does seem plausible that such a non-collinear 2D material may be
observed in the near future.

Again, bear in mind that all the calculations in the present notebook ought
to be properly converged with respect to $k$-points, plane wave cutoff etc.
"""



from ase import Atoms
from gpaw import GPAW, PW, FermiDirac

calc = GPAW(mode=PW(500),
            xc='PBE',
            h=0.10,
            occupations=FermiDirac(width=0.001),
            kpts={'size': (8, 8, 8), 'gamma': True},
            parallel={'band': 1, 'domain': 1},
            txt='gs_Bi2Se3.txt')

a = 4.138
c = 28.64
mu = 0.399
nu = 0.206
cell = [[-a / 2, -3**0.5 / 6 * a, c / 3],
        [a / 2, -3**0.5 / 6 * a, c / 3],
        [0.0, 3**0.5 / 3 * a, c / 3]]
pos = [[mu, mu, mu],
       [-mu, -mu, -mu],
       [0.0, 0.0, 0.0],
       [nu, nu, nu],
       [-nu, -nu, -nu]]
a = Atoms('Bi2Se3', cell=cell, scaled_positions=pos, pbc=True)
a.calc = calc
a.get_potential_energy()

calc.write('gs_Bi2Se3.gpw')



from gpaw.lrtddft import LrTDDFT, photoabsorption_spectrum

lr = LrTDDFT.read('Omega_Na2.gz')
lr.diagonalize()
lr.write('excitations_Na2.gz')

lr = LrTDDFT.read('excitations_Na2.gz')
photoabsorption_spectrum(lr, 'Na2_spectrum.dat', e_min=0.0, e_max=10)



import numpy as np
from functools import partial
from time import ctime

from ase.units import Hartree
from ase.utils import convert_string_to_fd
from ase.utils.timing import Timer, timer

from gpaw import extra_parameters
import gpaw.mpi as mpi
from gpaw.blacs import (BlacsGrid, BlacsDescriptor, Redistributor)
from gpaw.utilities.memory import maxrss
from gpaw.utilities.progressbar import ProgressBar
from gpaw.response.kspair import KohnShamPair, get_calc


class KohnShamLinearResponseFunction:
    r"""Class calculating linear response functions in the Kohn-Sham system

    Any linear response function can be calculated as a sum over transitions
    between the ground state and excited energy eigenstates.

    In the Kohn-Sham system this approach is particularly simple, as only
    excited states, for which a single electron has been moved from an occupied
    single-particle Kohn-Sham orbital to an unoccupied one, contribute.

    Resultantly, any linear response function in the Kohn-Sham system can be
    written as a sum over transitions between pairs of occupied and unoccupied
    Kohn-Sham orbitals.

    Currently, only collinear Kohn-Sham systems are supported. That is, all
    transitions can be written in terms of band indexes, k-points and spins:

    T (composit transition index): (n, k, s) -> (n', k', s')

    The sum over transitions is an integral over k-points in the 1st Brillouin
    Zone and a sum over all bands and spins. Sums over bands and spins can be
    handled together:

    t (composit transition index): (n, s) -> (n', s')
    
    __               __   __               __
    \      //        \    \      //        \
    /   =  ||dk dk'  /    /   =  ||dk dk'  /
    ‾‾     //        ‾‾   ‾‾     //        ‾‾
    T               n,n' s,s'              t
    """

    def __init__(self, gs, response=None, mode=None,
                 bandsummation='pairwise', nbands=None, kpointintegration=None,
                 world=mpi.world, nblocks=1, txt='-', timer=None):
        """Construct the KSLRF object

        Parameters
        ----------
        gs : str
            The groundstate calculation file that the linear response
            calculation is based on.
        response : str
            Type of response function.
            Currently, only susceptibilities are supported.
        mode: str
            Calculation mode.
            Currently, only a plane wave mode is implemented.
        bandsummation : str
            Band summation for pairs of Kohn-Sham orbitals
            'pairwise': sum over pairs of bands
            'double': double sum over band indices.
        nbands : int
            Maximum band index to include.
        kpointintegration : str
            Brillouin Zone integration for the Kohn-Sham orbital wave vector.
            Currently, only point integration is supported
        world : obj
            MPI communicator.
        nblocks : int
            Divide the response function storage into nblocks. Useful when the
            response function is large and memory requirements are restrictive.
        txt : str
            Output file.
        timer : func
            gpaw.utilities.timing.timer wrapper instance

        Attributes
        ----------
        kspair : gpaw.response.pair.KohnShamPair instance
            Class for handling pairs of Kohn-Sham orbitals
        pme : gpaw.response.pair.PairMatrixElement instance
            Class for calculating transition matrix elements for pairs of
            Kohn-Sham orbitals
        integrator : Integrator instance
            The integrator class is a general class for Brillouin Zone
            integration. The user defined integrand is integrated over k-points
            and summed over a given band and spin domain.

        Callables
        ---------
        self.add_integrand(kskptpair, weight, tmp_x, *args, **kwargs) : func
            Add the integrand for a given part of the domain to output array
        self.calculate(*args, **kwargs) : func
            Runs the calculation, returning the response function.
            Returned format can varry depending on response and mode.
        """
        # Output .txt filehandle
        self.fd = convert_string_to_fd(txt, world)
        self.cfd = self.fd
        print('Initializing KohnShamLinearResponseFunction', file=self.fd)

        # Communicators for parallelization
        self.world = world
        self.interblockcomm = None
        self.intrablockcomm = None
        self.initialize_communicators(nblocks)
        self.nblocks = self.interblockcomm.size

        # Timer
        self.timer = timer or Timer()

        # Load ground state calculation
        self.calc = get_calc(gs, fd=self.fd, timer=self.timer)

        # The KohnShamPair class handles data extraction from ground state
        self.kspair = KohnShamPair(self.calc, world=world,
                                   # Let each process handle slow steps only
                                   # for a fraction of all transitions.
                                   # t-transitions are distributed through
                                   # interblockcomm, k-points through
                                   # intrablockcomm.
                                   transitionblockscomm=self.interblockcomm,
                                   kptblockcomm=self.intrablockcomm,
                                   txt=self.fd, timer=self.timer)

        self.response = response
        self.mode = mode

        self.bandsummation = bandsummation
        self.nbands = nbands or self.calc.wfs.bd.nbands
        assert self.nbands <= self.calc.wfs.bd.nbands
        self.nocc1 = self.kspair.nocc1  # number of completely filled bands
        self.nocc2 = self.kspair.nocc2  # number of non-empty bands

        self.kpointintegration = kpointintegration
        self.integrator = create_integrator(self)
        # Each integrator might take some extra input kwargs
        self.extraintargs = {}

        # Attributes related to the specific response function
        self.pme = None

    def initialize_communicators(self, nblocks):
        """Set up MPI communicators to distribute the memory needed to store
        large arrays and parallelize calculations when possible.

        Parameters
        ----------
        nblocks : int
            Separate large arrays into n different blocks. Each process
            allocates memory for the large arrays. By allocating only a
            fraction/block of the total arrays, the memory requirements are
            eased.

        Sets
        ----
        interblockcomm : gpaw.mpi.Communicator
            Communicate between processes belonging to different memory blocks.
            In every communicator, there is one process for each block of
            memory, so that all blocks are represented.
            If nblocks < world.size, there will be size // nblocks different
            processes that allocate memory for the same block of the large
            arrays. Thus, there will be also size // nblocks different inter
            block communicators, grouping the processes into sets that allocate
            the entire arrays between them.
        intrablockcomm : gpaw.mpi.Communicator
            Communicate between processes belonging to the same memory block.
            There will be size // nblocks processes per memory block.
        """
        world = self.world
        if nblocks == 1:
            self.interblockcomm = world.new_communicator([world.rank])
            self.intrablockcomm = world
        else:
            assert world.size % nblocks == 0, world.size
            rank1 = world.rank // nblocks * nblocks
            rank2 = rank1 + nblocks
            self.interblockcomm = world.new_communicator(range(rank1, rank2))
            ranks = range(world.rank % nblocks, world.size, nblocks)
            self.intrablockcomm = world.new_communicator(ranks)
        print('Number of blocks:', nblocks, file=self.fd)

    @timer('Calculate Kohn-Sham linear response function')
    def calculate(self, spinrot=None, A_x=None):
        return self._calculate(spinrot, A_x)

    def _calculate(self, spinrot, A_x):
        """In-place calculation of the response function

        Parameters
        ----------
        spinrot : str
            Select spin rotation.
            Choices: 'u', 'd', '0' (= 'u' + 'd'), '-'= and '+'
            All rotations are included for spinrot=None ('0' + '+' + '-').
        A_x : ndarray
            Output array. If None, the output array is created.
        """
        self.spinrot = spinrot
        # Prepare to sum over bands and spins
        n1_t, n2_t, s1_t, s2_t = self.get_band_spin_transitions_domain()

        # Print information about the prepared calculation
        self.print_information(len(n1_t))
        if extra_parameters.get('df_dry_run'):  # Exit after setting up
            print('    Dry run exit', file=self.fd)
            raise SystemExit

        A_x = self.setup_output_array(A_x)

        self.integrator.integrate(n1_t, n2_t, s1_t, s2_t,
                                  out_x=A_x, **self.extraintargs)

        # Different calculation modes might want the response function output
        # in different formats
        out = self.post_process(A_x)

        print('', file=self.cfd)

        return out

    def get_band_spin_transitions_domain(self):
        """Generate all allowed band and spin transitions.
        
        If only a subset of possible spin rotations are considered
        (examples: s1 = s2 or s2 = 1 - s1), do not include others
        in the sum over transitions.
        """
        n1_M, n2_M = get_band_transitions_domain(self.bandsummation,
                                                 self.nbands,
                                                 nocc1=self.nocc1,
                                                 nocc2=self.nocc2)
        s1_S, s2_S = get_spin_transitions_domain(self.bandsummation,
                                                 self.spinrot,
                                                 self.calc.wfs.nspins)

        return transitions_in_composite_index(n1_M, n2_M, s1_S, s2_S)

    def setup_output_array(self, A_x):
        raise NotImplementedError('Output array depends on mode')

    def get_ks_kpoint_pairs(self, k_pv, *args, **kwargs):
        raise NotImplementedError('Integrated pairs of states depend on'
                                  'response and mode')

    def initialize_pme(self, *args, **kwargs):
        raise NotImplementedError('Calculator method for matrix elements '
                                  'depend on response and mode')

    def calculate_pme(self, kskptpair, *args, **kwargs):
        raise NotImplementedError('Calculator method for matrix elements '
                                  'depend on response and mode')

    def add_integrand(self, kskptpair, weight, tmp_x, *args, **kwargs):
        raise NotImplementedError('Integrand depends on response and mode')

    def post_process(self, A_x):
        raise NotImplementedError('Post processing depends on mode')

    def print_information(self, nt):
        """Basic information about the input ground state, parallelization
        and sum over states"""
        ns = self.calc.wfs.nspins
        nbands = self.nbands
        nocc = self.nocc1
        npocc = self.nocc2
        nk = self.calc.wfs.kd.nbzkpts
        nik = self.calc.wfs.kd.nibzkpts

        if extra_parameters.get('df_dry_run'):
            from gpaw.mpi import DryRunCommunicator
            size = extra_parameters['df_dry_run']
            world = DryRunCommunicator(size)
        else:
            world = self.world
        wsize = world.size
        knsize = self.intrablockcomm.size
        bsize = self.interblockcomm.size

        spinrot = self.spinrot

        p = partial(print, file=self.cfd)

        p('Called a response.kslrf.KohnShamLinearResponseFunction.calculate()')
        p('%s' % ctime())
        p('Using a Kohn-Sham ground state with:')
        p('    Number of spins: %d' % ns)
        p('    Number of bands: %d' % nbands)
        p('    Number of completely occupied states: %d' % nocc)
        p('    Number of partially occupied states: %d' % npocc)
        p('    Number of kpoints: %d' % nk)
        p('    Number of irredicible kpoints: %d' % nik)
        p('')
        p('The response function calculation is performed in parallel with:')
        p('    world.size: %d' % wsize)
        p('    intrablockcomm.size: %d' % knsize)
        p('    interblockcomm.size: %d' % bsize)
        p('')
        p('The sum over band and spin transitions is performed using:')
        p('    Spin rotation: %s' % spinrot)
        p('    Total number of composite band and spin transitions: %d' % nt)
        p('')


def get_band_transitions_domain(bandsummation, nbands, nocc1=None, nocc2=None):
    """Get all pairs of bands to sum over

    Parameters
    ----------
    bandsummation : str
        Band summation method
    nbands : int
        number of bands
    nocc1 : int
        number of completely filled bands
    nocc2 : int
        number of non-empty bands

    Returns
    -------
    n1_M : ndarray
        band index 1, M = (n1, n2) composite index
    n2_M : ndarray
        band index 2, M = (n1, n2) composite index
    """
    _get_band_transitions_domain =\
        create_get_band_transitions_domain(bandsummation)
    n1_M, n2_M = _get_band_transitions_domain(nbands)
    
    return remove_null_transitions(n1_M, n2_M, nocc1=nocc1, nocc2=nocc2)


def create_get_band_transitions_domain(bandsummation):
    """Creator component deciding how to carry out band summation."""
    if bandsummation == 'pairwise':
        return get_pairwise_band_transitions_domain
    elif bandsummation == 'double':
        return get_double_band_transitions_domain
    raise ValueError(bandsummation)


def get_double_band_transitions_domain(nbands):
    """Make a simple double sum"""
    n_n = np.arange(0, nbands)
    m_m = np.arange(0, nbands)
    n_nm, m_nm = np.meshgrid(n_n, m_m)
    n_M, m_M = n_nm.flatten(), m_nm.flatten()

    return n_M, m_M


def get_pairwise_band_transitions_domain(nbands):
    """Make a sum over all pairs"""
    n_n = range(0, nbands)
    n_M = []
    m_M = []
    for n in n_n:
        m_m = range(n, nbands)
        n_M += [n] * len(m_m)
        m_M += m_m

    return np.array(n_M), np.array(m_M)


def remove_null_transitions(n1_M, n2_M, nocc1=None, nocc2=None):
    """Remove pairs of bands, between which transitions are impossible"""
    n1_newM = []
    n2_newM = []
    for n1, n2 in zip(n1_M, n2_M):
        if nocc1 is not None and (n1 < nocc1 and n2 < nocc1):
            continue  # both bands are fully occupied
        elif nocc2 is not None and (n1 >= nocc2 and n2 >= nocc2):
            continue  # both bands are completely unoccupied
        n1_newM.append(n1)
        n2_newM.append(n2)

    return np.array(n1_newM), np.array(n2_newM)


def get_spin_transitions_domain(bandsummation, spinrot, nspins):
    """Get structure of the sum over spins

    Parameters
    ----------
    bandsummation : str
        Band summation method
    spinrot : str
        spin rotation
    nspins : int
        number of spin channels in ground state calculation

    Returns
    -------
    s1_s : ndarray
        spin index 1, S = (s1, s2) composite index
    s2_S : ndarray
        spin index 2, S = (s1, s2) composite index
    """
    _get_spin_transitions_domain =\
        create_get_spin_transitions_domain(bandsummation)
    return _get_spin_transitions_domain(spinrot, nspins)


def create_get_spin_transitions_domain(bandsummation):
    """Creator component deciding how to carry out spin summation."""
    if bandsummation == 'pairwise':
        return get_pairwise_spin_transitions_domain
    elif bandsummation == 'double':
        return get_double_spin_transitions_domain
    raise ValueError(bandsummation)


def get_double_spin_transitions_domain(spinrot, nspins):
    """Usual spin rotations forward in time"""
    if nspins == 1:
        if spinrot is None or spinrot == '0':
            s1_S = [0]
            s2_S = [0]
        else:
            raise ValueError(spinrot, nspins)
    else:
        if spinrot is None:
            s1_S = [0, 0, 1, 1]
            s2_S = [0, 1, 0, 1]
        elif spinrot == '0':
            s1_S = [0, 1]
            s2_S = [0, 1]
        elif spinrot == 'u':
            s1_S = [0]
            s2_S = [0]
        elif spinrot == 'd':
            s1_S = [1]
            s2_S = [1]
        elif spinrot == '-':
            s1_S = [0]  # spin up
            s2_S = [1]  # spin down
        elif spinrot == '+':
            s1_S = [1]  # spin down
            s2_S = [0]  # spin up
        else:
            raise ValueError(spinrot)

    return np.array(s1_S), np.array(s2_S)


def get_pairwise_spin_transitions_domain(spinrot, nspins):
    """In a sum over pairs, transitions including a spin rotation may have to
    include terms, propagating backwards in time."""
    if spinrot in ['+', '-']:
        assert nspins == 2
        return np.array([0, 1]), np.array([1, 0])
    else:
        return get_double_spin_transitions_domain(spinrot, nspins)


def transitions_in_composite_index(n1_M, n2_M, s1_S, s2_S):
    """Use a composite index t for transitions (n, s) -> (n', s')."""
    n1_MS, s1_MS = np.meshgrid(n1_M, s1_S)
    n2_MS, s2_MS = np.meshgrid(n2_M, s2_S)
    return n1_MS.flatten(), n2_MS.flatten(), s1_MS.flatten(), s2_MS.flatten()


class PlaneWaveKSLRF(KohnShamLinearResponseFunction):
    """Class for doing KS-LRF calculations in plane wave mode"""

    def __init__(self, *args, eta=0.2, ecut=50, gammacentered=False,
                 disable_point_group=True, disable_time_reversal=True,
                 disable_non_symmorphic=True, bundle_integrals=True,
                 kpointintegration='point integration', bundle_kptpairs=False,
                 **kwargs):
        """Initialize the plane wave calculator mode.
        In plane wave mode, the linear response function is calculated for a
        given set of frequencies. The spatial part is expanded in plane waves
        for a given momentum transfer q within the first Brillouin Zone.

        Parameters
        ----------
        eta : float
            Energy broadening of spectra.
        ecut : float
            Energy cutoff for the plane wave representation.
        gammacentered : bool
            Center the grid of plane waves around the gamma point or q-vector.
        disable_point_group : bool
            Do not use the point group symmetry operators.
        disable_time_reversal : bool
            Do not use time reversal symmetry.
        disable_non_symmorphic : bool
            Do no use non symmorphic symmetry operators.
        bundle_integrals : bool
            Do the k-point integrals (large matrix multiplications)
            simultaneously for all frequencies.
            Can be switched of, if this step forces calculations out of memory.
        bundle_kptpairs : bool
            Extract the k-point pairs simultaneously, so no process has to wait
            for the others in the middle of the response function integration.
            [Only relevant in the case where ground state is distributed]
            Can be switched of, if this step forces calculations out of memory.
        """
        from gpaw.mpi import SerialCommunicator

        # Avoid any mode ambiguity
        if 'mode' in kwargs.keys():
            mode = kwargs.pop('mode')
            assert mode == 'pw'

        KSLRF = KohnShamLinearResponseFunction
        KSLRF.__init__(self, *args, mode='pw',
                       kpointintegration=kpointintegration, **kwargs)

        self.eta = eta / Hartree
        self.ecut = None if ecut is None else ecut / Hartree
        self.gammacentered = gammacentered

        self.disable_point_group = disable_point_group
        self.disable_time_reversal = disable_time_reversal
        self.disable_non_symmorphic = disable_non_symmorphic

        self.bundle_integrals = bundle_integrals

        # Bundle kptpairs if specified and using a distributed ground state
        cworld = self.calc.world
        if isinstance(cworld, SerialCommunicator) or cworld.size == 1:
            bundle_kptpairs = False
        self.bundle_kptpairs = bundle_kptpairs

        # Attributes related to specific q, given to self.calculate()
        self.pd = None  # Plane wave descriptor for given momentum transfer q
        self.pwsa = None  # Plane wave symmetry analyzer for given q
        self.wd = None  # Frequency descriptor for the given frequencies
        self.omega_w = None  # Frequencies in code units

    @timer('Calculate Kohn-Sham linear response function in plane wave mode')
    def calculate(self, q_c, frequencies, spinrot=None, A_x=None):
        """
        Parameters
        ----------
        q_c : list or ndarray or PWDescriptor
            Momentum transfer (and possibly plane wave basis)
        frequencies : ndarray or FrequencyDescriptor
            Array of frequencies to evaluate the response function at or
            descriptor of those frequencies.

        Returns
        -------
        pd : Planewave descriptor
            Planewave descriptor for q_c.
        A_wGG : ndarray
            The linear response function.
        """
        # Set up plane wave description with the gived momentum transfer q
        self.pd = self.get_PWDescriptor(q_c)
        self.pwsa = self.get_PWSymmetryAnalyzer(self.pd)

        # Set up frequency descriptor for the given frequencies
        self.wd = self.get_FreqDescriptor(frequencies)
        self.omega_w = self.wd.get_data()

        # In-place calculation
        return self._calculate(spinrot, A_x)

    def get_PWDescriptor(self, q_c):
        """Get the planewave descriptor for a certain momentum transfer q_c."""
        from gpaw.wavefunctions.pw import PWDescriptor
        if isinstance(q_c, PWDescriptor):
            return q_c
        else:
            from gpaw.kpt_descriptor import KPointDescriptor
            q_c = np.asarray(q_c, dtype=float)
            qd = KPointDescriptor([q_c])
            pd = PWDescriptor(self.ecut, self.calc.wfs.gd,
                              complex, qd, gammacentered=self.gammacentered)
            return pd

    @timer('Get PW symmetry analyser')
    def get_PWSymmetryAnalyzer(self, pd):
        from gpaw.response.pair import PWSymmetryAnalyzer as PWSA
        
        return PWSA(self.calc.wfs.kd, pd,
                    timer=self.timer, txt=self.fd,
                    disable_point_group=self.disable_point_group,
                    disable_time_reversal=self.disable_time_reversal,
                    disable_non_symmorphic=self.disable_non_symmorphic)

    def get_FreqDescriptor(self, frequencies):
        """Get the frequency descriptor for a certain input frequencies."""
        if isinstance(frequencies, FrequencyDescriptor):
            return frequencies
        else:
            return FrequencyDescriptor(np.asarray(frequencies) / Hartree)

    def print_information(self, nt):
        """Basic information about the input ground state, parallelization,
        sum over states and calculated response function array."""
        KohnShamLinearResponseFunction.print_information(self, nt)

        pd = self.pd
        q_c = pd.kd.bzk_kc[0]
        nw = len(self.omega_w)
        eta = self.eta * Hartree
        ecut = self.ecut * Hartree
        ngmax = pd.ngmax
        Asize = nw * pd.ngmax**2 * 16. / 1024**2 / self.interblockcomm.size

        p = partial(print, file=self.cfd)

        p('The response function is calculated in the PlaneWave mode, using:')
        p('    q_c: [%f, %f, %f]' % (q_c[0], q_c[1], q_c[2]))
        p('    Number of frequency points: %d' % nw)
        p('    Broadening (eta): %f' % eta)
        p('    Planewave cutoff: %f' % ecut)
        p('    Number of planewaves: %d' % ngmax)
        p('')
        p('    Memory estimates:')
        p('        A_wGG: %f M / cpu' % Asize)
        p('        Memory usage before allocation: %f M / cpu' % (maxrss() /
                                                                  1024**2))
        p('')

    def setup_output_array(self, A_x=None):
        """Initialize the output array in blocks"""
        # Could use some more documentation XXX
        nG = self.pd.ngmax
        nw = len(self.omega_w)
        mynG = (nG + self.interblockcomm.size - 1) // self.interblockcomm.size
        self.Ga = min(self.interblockcomm.rank * mynG, nG)
        self.Gb = min(self.Ga + mynG, nG)
        # if self.interblockcomm.rank == 0:
        #     assert self.Gb - self.Ga >= 3
        # assert mynG * (self.interblockcomm.size - 1) < nG
        if self.bundle_integrals:
            # Setup A_GwG
            if A_x is not None:
                nx = nw * (self.Gb - self.Ga) * nG
                A_GwG = A_x[:nx].reshape((nG, nw, self.Gb - self.Ga))
                A_GwG[:] = 0.0
            else:
                A_GwG = np.zeros((nG, nw, self.Gb - self.Ga), complex)

            return A_GwG
        else:
            # Setup A_wGG
            if A_x is not None:
                nx = nw * (self.Gb - self.Ga) * nG
                A_wGG = A_x[:nx].reshape((nw, self.Gb - self.Ga, nG))
                A_wGG[:] = 0.0
            else:
                A_wGG = np.zeros((nw, self.Gb - self.Ga, nG), complex)

            return A_wGG

    def get_ks_kpoint_pairs(self, k_pv, n1_t, n2_t, s1_t, s2_t):
        """Get all pairs of Kohn-Sham transitions:

        (n1_t, k_c, s1_t) -> (n2_t, k_c + q_c, s2_t)

        for each process with its own k-point.
        """
        k_pc = np.array([np.dot(self.pd.gd.cell_cv, k_v) / (2 * np.pi)
                         for k_v in k_pv])
        q_c = self.pd.kd.bzk_kc[0]
        return self.kspair.get_kpoint_pairs(n1_t, n2_t, k_pc, k_pc + q_c,
                                            s1_t, s2_t)

    def initialize_pme(self):
        self.pme.initialize(self.pd)

    def calculate_pme(self, kskptpair):
        self.pme(kskptpair, self.pd)

    def add_integrand(self, kskptpair, weight, tmp_x, **kwargs):
        raise NotImplementedError('Integrand depends on response')

    @timer('Post processing')
    def post_process(self, A_x):
        if self.bundle_integrals:
            # A_x = A_GwG
            A_wGG = A_x.transpose((1, 2, 0))
        else:
            A_wGG = A_x

        tmpA_wGG = self.redistribute(A_wGG)  # distribute over frequencies
        with self.timer('Symmetrizing Kohn-Sham linear response function'):
            self.pwsa.symmetrize_wGG(tmpA_wGG)
        self.redistribute(tmpA_wGG, A_wGG)

        return self.pd, A_wGG

    @timer('Redistribute memory')
    def redistribute(self, in_wGG, out_x=None):
        """Redistribute array.

        Switch between two kinds of parallel distributions:

        1) parallel over G-vectors (second dimension of in_wGG)
        2) parallel over frequency (first dimension of in_wGG)

        Returns new array using the memory in the 1-d array out_x.
        """

        comm = self.interblockcomm

        if comm.size == 1:
            return in_wGG

        nw = len(self.omega_w)
        nG = in_wGG.shape[2]
        mynw = (nw + comm.size - 1) // comm.size
        mynG = (nG + comm.size - 1) // comm.size

        bg1 = BlacsGrid(comm, comm.size, 1)
        bg2 = BlacsGrid(comm, 1, comm.size)
        md1 = BlacsDescriptor(bg1, nw, nG**2, mynw, nG**2)
        md2 = BlacsDescriptor(bg2, nw, nG**2, nw, mynG * nG)

        if len(in_wGG) == nw:
            mdin = md2
            mdout = md1
        else:
            mdin = md1
            mdout = md2

        r = Redistributor(comm, mdin, mdout)

        outshape = (mdout.shape[0], mdout.shape[1] // nG, nG)
        if out_x is None:
            out_wGG = np.empty(outshape, complex)
        else:
            out_wGG = out_x[:np.product(outshape)].reshape(outshape)

        r.redistribute(in_wGG.reshape(mdin.shape),
                       out_wGG.reshape(mdout.shape))

        return out_wGG


class FrequencyDescriptor:
    """Describes a one-dimensional array of frequencies."""

    def __init__(self, data_x):
        self.data_x = np.array(np.sort(data_x))
        self._data_len = len(data_x)

    def __len__(self):
        return self._data_len

    def get_data(self):
        return self.data_x


class Integrator:
    """Baseclass for integrating over k-points in the first Brillouin Zone
    and summing over bands and spin.
    """
    def __init__(self, kslrf):
        """
        Parameters
        ----------
        kslrf : KohnShamLinearResponseFunction instance
        """
        self.kslrf = kslrf
        self.timer = self.kslrf.timer

    def slice_kpoint_domain(self, bzk_kv, weight_k):
        """When integrating over k-points, slice the domain in pieces with one
        k-point per process each.

        Returns
        -------
        bzk_ipv : nd.array
            k-points coordinates for each process for each iteration
        """
        nk = bzk_kv.shape[0]
        size = self.kslrf.intrablockcomm.size
        ni = (nk + size - 1) // size
        bzk_ipv = np.array([bzk_kv[i * size:(i + 1) * size]
                            for i in range(ni)])

        # Extract the weight corresponding to the process' own k-point pair
        weight_ip = np.array([weight_k[i * size:(i + 1) * size]
                              for i in range(ni)])
        weight_i = [None] * len(weight_ip)
        krank = self.kslrf.intrablockcomm.rank
        for i, w_p in enumerate(weight_ip):
            if krank in range(len(w_p)):
                weight_i[i] = w_p[krank]

        return bzk_ipv, weight_i

    @timer('Integrate response function')
    def integrate(self, n1_t, n2_t, s1_t, s2_t,
                  out_x=None, **kwargs):
        if out_x is None:
            raise NotImplementedError

        bzk_kv, weight_k = self.get_kpoint_domain()
        prefactor = self.calculate_bzint_prefactor(bzk_kv)
        out_x /= prefactor
        self._integrate(bzk_kv, weight_k,
                        n1_t, n2_t, s1_t, s2_t, out_x, **kwargs)
        out_x *= prefactor
        
        return out_x

    def get_kpoint_domain(self):
        raise NotImplementedError('Domain depends on integration method')

    def calculate_bzint_prefactor(self, bzk_kv):
        raise NotImplementedError('Prefactor depends on integration method')

    def _integrate(self, bzk_kv, weight_k,
                   n1_t, n2_t, s1_t, s2_t, out_x, **kwargs):
        raise NotImplementedError('Integration method is defined by subclass')


class PWPointIntegrator(Integrator):
    """A simple point integrator for the plane wave mode."""

    @timer('Get k-point domain')
    def get_kpoint_domain(self):
        # Could use some more documentation XXX
        K_gK = self.kslrf.pwsa.group_kpoints()
        bzk_kc = np.array([self.kslrf.calc.wfs.kd.bzk_kc[K_K[0]] for
                           K_K in K_gK])

        bzk_kv = np.dot(bzk_kc, self.kslrf.pd.gd.icell_cv) * 2 * np.pi

        nsym = self.kslrf.pwsa.how_many_symmetries()
        weight_k = [self.kslrf.pwsa.get_kpoint_weight(k_c) / nsym
                    for k_c in bzk_kc]

        return bzk_kv, weight_k

    def calculate_bzint_prefactor(self, bzk_kv):
        # Could use some more documentation XXX
        if self.kslrf.calc.wfs.kd.refine_info is not None:
            nbzkpts = self.kslrf.calc.wfs.kd.refine_info.mhnbzkpts
        else:
            nbzkpts = self.kslrf.calc.wfs.kd.nbzkpts
        frac = len(bzk_kv) / nbzkpts
        
        return (2 * frac * self.kslrf.pwsa.how_many_symmetries() /
                (self.kslrf.calc.wfs.nspins * (2 * np.pi)**3))

    def _integrate(self, bzk_kv, weight_k,
                   n1_t, n2_t, s1_t, s2_t, out_x, **kwargs):
        """Do a simple sum over k-points in the first Brillouin Zone,
        adding the integrand (summed over bands and spin) for each k-point."""

        nk = bzk_kv.shape[0]
        vol = abs(np.linalg.det(self.kslrf.calc.wfs.gd.cell_cv))

        kpointvol = (2 * np.pi)**3 / vol / nk
        out_x /= kpointvol

        # Initialize pme
        print('----------',
              file=self.kslrf.cfd)
        print('Initializing PairMatrixElement',
              file=self.kslrf.cfd, flush=True)
        self.kslrf.initialize_pme()

        # Slice domain
        bzk_ipv, weight_i = self.slice_kpoint_domain(bzk_kv, weight_k)

        # Perform sum over k-points
        tmp_x = np.zeros_like(out_x)
        pb = ProgressBar(self.kslrf.cfd)
        if self.kslrf.bundle_kptpairs:
            # Extract all the process' kptpairs at once, then do integration
            print('\nExtracting Kohn-Sham k-point pairs',
                  file=self.kslrf.cfd, flush=True)
            kskptpair_i = []
            for k_pv in bzk_ipv:
                kskptpair_i.append(self.kslrf.get_ks_kpoint_pairs(k_pv,
                                                                  n1_t, n2_t,
                                                                  s1_t, s2_t))

            print('\nIntegrating response function',
                  file=self.kslrf.cfd, flush=True)
            # Carry out sum
            for i, kskptpair in pb.enumerate(kskptpair_i):
                if kskptpair is not None:
                    weight = weight_i[i]
                    assert weight is not None
                    self.kslrf.calculate_pme(kskptpair)
                    self.kslrf.add_integrand(kskptpair, weight,
                                             tmp_x, **kwargs)
        else:
            # Each process will do its own k-points, but it has to follow the
            # others, as it may have to send them information about its
            # partition of the ground state
            print('\nIntegrating response function',
                  file=self.kslrf.cfd, flush=True)
            for i, k_pv in pb.enumerate(bzk_ipv):
                kskptpair = self.kslrf.get_ks_kpoint_pairs(k_pv, n1_t, n2_t,
                                                           s1_t, s2_t)
                if kskptpair is not None:
                    weight = weight_i[i]
                    assert weight is not None
                    self.kslrf.calculate_pme(kskptpair)
                    self.kslrf.add_integrand(kskptpair, weight,
                                             tmp_x, **kwargs)

        # Sum over the k-points that have been distributed between processes
        with self.timer('Sum over distributed k-points'):
            self.kslrf.intrablockcomm.sum(tmp_x)

        out_x += tmp_x
        out_x *= kpointvol


def create_integrator(kslrf):
    """Creator component for the integrator"""
    if kslrf.mode == 'pw':
        if kslrf.kpointintegration is None or \
           kslrf.kpointintegration == 'point integration':
            return PWPointIntegrator(kslrf)

    raise ValueError(kslrf.mode, kslrf.kpointintegration)



# Creates: nl_freq_grid.png
import numpy as np
import matplotlib.pyplot as plt

from gpaw.response.chi0 import frequency_grid


omegamax = 50.0
domega0 = 0.2

plt.figure(figsize=(5, 5))
for omega2 in [2.5, 5, 10, 15, 20, np.inf]:
    x = frequency_grid(domega0, omega2, omegamax)
    y = range(len(x))
    if omega2 == np.inf:
        label = '$\\omega_2 = \\infty$'
    else:
        label = '$\\omega_2 = %.1f\\, \\mathrm{eV}$' % omega2
    plt.plot(x, y, '.', label=label)
plt.ylabel('Freq. no')
plt.xlabel(r'$\omega\, [\mathrm{eV}]$')
plt.axis(xmax=30, ymax=200)
plt.title(r'$\Delta\omega_0 = 0.2\, \mathrm{eV}$')
plt.legend(loc=2)
plt.savefig('nl_freq_grid.png', bbox_inches='tight')



'Test ase.dft.wannier module with k-points.'

import pytest
from ase.build import bulk
from ase.dft.wannier import Wannier

from gpaw import GPAW
from gpaw.mpi import world, serial_comm


@pytest.mark.skip(reason='TODO')
def test_wannierk_lcao():
    si = bulk('Si', 'diamond', a=5.43)

    k = 2
    if 1:
        si.calc = GPAW(kpts=(k, k, k), mode='lcao', txt='Si-ibz.txt')
        # si.calc = GPAW(kpts=(k, k, k), mode='lcao')
        _ = si.get_potential_energy()
        si.calc.write('Si-ibz', mode='all')
        # si.calc.set(symmetry='off', txt='Si-bz.txt')
        # si.calc.set(symmetry='off')
        # e2 = si.get_potential_energy()
        # si.calc.write('Si-bz', mode='all')
        # print e1, e2

    def wan(calc):
        centers = [([0.125, 0.125, 0.125], 0, 1.5),
                   ([0.125, 0.625, 0.125], 0, 1.5),
                   ([0.125, 0.125, 0.625], 0, 1.5),
                   ([0.625, 0.125, 0.125], 0, 1.5)]
        w = Wannier(4, calc,
                    nbands=4,
                    verbose=False,
                    initialwannier=centers)
        w.localize()
        x = w.get_functional_value()
        centers = (w.get_centers(1) * k) % 1
        c = (centers - 0.125) * 2
        # print w.get_radii()  # broken! XXX
        assert abs(c.round() - c).max() < 0.03
        c = sorted(c.round().astype(int).tolist())
        assert c == [[0, 0, 0], [0, 0, 1], [0, 1, 0], [1, 0, 0]]
        if 0:
            from ase.visualize import view
            from ase import Atoms
            watoms = calc.atoms + Atoms(symbols='X4',
                                        scaled_positions=centers,
                                        cell=calc.atoms.cell)
            view(watoms)
        return x

    calc1 = GPAW('Si-bz.gpw', txt=None, communicator=serial_comm)
    # do full lcao initialization
    # print('Re-Init LCAO calc')
    _s = calc1.get_atoms()
    calc1.set_positions(_s)
    calc1.wfs.eigensolver.iterate(calc1.hamiltonian, calc1.wfs)
    x1 = wan(calc1)

    # print x1
    x2 = 5.76962685681      # from FD mode
    assert abs(x1 - x2) < 0.001
    # assert abs(x1 - 9.71) < 0.01

    world.barrier()




from ase import Atoms
from gpaw import GPAW
from gpaw.test import equal
from gpaw.wavefunctions.pw import PW

from gpaw.external import ConstantPotential


def test_ext_potential_external_pw():
    ConstantPotential()

    sc = 2.9
    R = 0.7  # approx. experimental bond length
    R = 1.0
    a = 2 * sc
    c = 3 * sc
    H2 = Atoms('HH', [(a / 2, a / 2, (c - R) / 2),
                      (a / 2, a / 2, (c + R) / 2)],
               cell=(a, a, c),
               pbc=False)

    txt = None

    convergence = {'eigenstates': 1.e-6 * 40 * 1.5**3,
                   'density': 1.e-2,
                   'energy': 0.1}

    # without potential
    if True:
        if txt:
            print('\n################## no potential')
        c00 = GPAW(mode=PW(200), nbands=-1,
                   convergence=convergence,
                   txt=txt)
        c00.calculate(H2)
        c00.get_eigenvalues()

    # 0 potential
    if True:
        if txt:
            print('\n################## 0 potential')
        cp0 = ConstantPotential(0.0)
        c01 = GPAW(mode=PW(200), nbands=-2, external=cp0,
                   convergence=convergence,
                   txt=txt)
        c01.calculate(H2)

    # 1 potential
    if True:
        if txt:
            print('################## 1 potential')
        cp1 = ConstantPotential(-1.0)
        c1 = GPAW(mode=PW(200), nbands=-2, external=cp1,
                  convergence=convergence,
                  txt=txt)
        c1.calculate(H2)

    for i in range(c00.get_number_of_bands()):
        f00 = c00.get_occupation_numbers()[i]
        if f00 > 0.01:
            e00 = c00.get_eigenvalues()[i]
            e1 = c1.get_eigenvalues()[i]
            print('Eigenvalues no pot, expected, error=',
                  e00, e1 + 1, e00 - e1 - 1)
            equal(e00, e1 + 1., 0.02)

    E_c00 = c00.get_potential_energy()
    E_c1 = c1.get_potential_energy()
    DeltaE = E_c00 - E_c1
    equal(DeltaE, 0, 0.002)



from ase.build import fcc111, add_adsorbate
from gpaw import GPAW, PW

#  Slab with CO:
slab = fcc111('Pt', size=(1, 1, 3))
add_adsorbate(slab, 'C', 2.0, 'ontop')
add_adsorbate(slab, 'O', 3.15, 'ontop')
slab.center(axis=2, vacuum=4.0)
slab.calc = GPAW(mode=PW(400),
                 xc='RPBE',
                 kpts=(12, 12, 1),
                 convergence={'bands': -10},
                 txt='top.txt')
slab.get_potential_energy()
slab.calc.write('top.gpw', mode='all')

#  Molecule
molecule = slab[-2:]
molecule.calc = GPAW(mode=PW(400),
                     xc='RPBE',
                     kpts=(12, 12, 1),
                     txt='CO.txt')

molecule.get_potential_energy()
molecule.calc.write('CO.gpw', mode='all')
